# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
NSIS Archive Engine Plugin

This plugin handles NSIS (Nullsoft Scriptable Install System) format for scanning and manipulation.
"""

import contextlib
import datetime
import logging
import mmap
import os
import struct
import tempfile
import zlib
from ctypes import *
from io import BytesIO

try:
    import pylzma

    HAS_PYLZMA = True
except ImportError:
    HAS_PYLZMA = False

from kicomav.plugins import cryptolib
from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------
# Structure for NSIS
# ----------------------------------------------------------------------------
BYTE = c_ubyte
WORD = c_ushort
LONG = c_long
DWORD = c_uint
FLOAT = c_float
LPBYTE = POINTER(c_ubyte)
LPTSTR = POINTER(c_char)
HANDLE = c_void_p
PVOID = c_void_p
LPVOID = c_void_p
UINT_PTR = c_uint
SIZE_T = c_uint


class StructNsisHeader(Structure):
    _pack_ = 1
    _fields_ = [
        ("flag", DWORD),
        ("pages", DWORD),
        ("pages_num", DWORD),
        ("sections", DWORD),
        ("sections_num", DWORD),
        ("entries", DWORD),
        ("entries_num", DWORD),
        ("strings", DWORD),
        ("strings_num", DWORD),
        ("langtables", DWORD),
        ("langtables_num", DWORD),
        ("ctlcolors", DWORD),
        ("ctlcolors_num", DWORD),
        ("bgfont", DWORD),
        ("bgfont_num", DWORD),
        ("data", DWORD),
        ("data_num", DWORD),
    ]


class StructNsisRecord(Structure):
    _pack_ = 1
    _fields_ = [
        ("which", DWORD),
        ("parm0", LONG),
        ("parm1", LONG),
        ("parm2", LONG),
        ("parm3", LONG),
        ("parm4", LONG),
        ("parm5", LONG),
    ]


NsisVarNames = {
    # init with 1
    0: "0",
    1: "1",
    2: "2",
    3: "3",
    4: "4",
    5: "5",
    6: "6",
    7: "7",
    8: "8",
    9: "9",
    10: "R0",
    11: "R1",
    12: "R2",
    13: "R3",
    14: "R4",
    15: "R5",
    16: "R6",
    17: "R7",
    18: "R8",
    19: "R9",
    20: "CMDLINE",
    21: "INSTDIR",
    22: "OUTDIR",
    23: "EXEDIR",
    24: "LANGUAGE",
    # init with -1
    25: "TEMP",
    26: "PLUGINSDIR",
    27: "EXEPATH",
    28: "EXEFILE",
    29: "HWNDPARENT",
    30: "_CLICK",
    # init with 1
    31: "_OUTDIR",
}


class NSIS:
    TYPE_UNKNOWN = -1
    TYPE_LZMA = 0
    TYPE_BZIP2 = 1
    TYPE_ZLIB = 2
    TYPE_COPY = 3

    def __init__(self, filename, offset=0, verbose=False):
        self.verbose = verbose
        self.filename = filename
        self.fp = None
        self.mm = None

        self.nsis_header = None
        self.body_data = None
        self.case_type = 0

        self.temp_name = None
        self.start_offset = offset

    def parse(self):
        # Use mkstemp instead of mktemp to prevent race condition (CWE-377)
        fd, self.temp_name = tempfile.mkstemp(prefix="knsf")

        with open(self.filename, "rb") as fp:
            fp.seek(self.start_offset)
            data = fp.read()

        try:
            os.write(fd, data)
        finally:
            os.close(fd)

        self.fp = open(self.temp_name, "rb")
        fsize = os.path.getsize(self.temp_name)
        if fsize == 0:
            return False

        self.mm = mmap.mmap(self.fp.fileno(), 0, access=mmap.ACCESS_READ)

        flag = kavutil.get_uint32(self.mm, 0)
        head_size = kavutil.get_uint32(self.mm, 0x14)
        comp_size = kavutil.get_uint32(self.mm, 0x18)

        data, case_type = self.get_data()  # Get all NSIS data
        self.body_data = data
        self.case_type = case_type

        if self.verbose:
            self.print_nsis_debug_info(flag, case_type, data)
        return True

    def print_nsis_debug_info(self, flag, case_type, data):
        """
        Prints detailed debug information about NSIS file format

        Args:
            flag: NSIS header flag
            case_type: Compression case type
            data: Uncompressed NSIS data

        Prints:
            - Engine information
            - NSIS header details
            - Uncompressed data preview
            - File extraction information
        """
        print("-" * 79)
        kavutil.vprint("Engine")
        kavutil.vprint(None, "Engine", "nsis")
        kavutil.vprint(None, "File name", os.path.split(self.filename)[-1])

        print()
        kavutil.vprint("NSIS")
        kavutil.vprint(None, "Flag", f"{flag}")
        kavutil.vprint(None, "Uncompress Case", f"{case_type}")

        self.print_section_header("Uncompress Data")
        kavutil.HexDump().Buffer(data, 0, 0x80)

        s = self.nsis_header.namelist_ex()
        if len(s):
            self.print_section_header("File Extract")
            for t in s:
                (foff, fname, ftime, extract_type) = t
                print("%08X | %-45s | %s" % (foff, fname, ftime if ftime != "" else "N/A"))

    def print_section_header(self, section_name):
        """
        Prints a formatted section header with empty lines

        Args:
            section_name: Name of the section to display
        """
        print()
        kavutil.vprint(section_name)
        print()

    def namelist(self):
        return self.nsis_header.namelist()

    def read(self, filename):
        if filename not in self.nsis_header.files:
            return None
        data = None
        (foff, ftime, extract_type) = self.nsis_header.files[filename]

        if self.case_type == 1:  # case 1: Compressing the entire installation file
            fsize = kavutil.get_uint32(self.body_data, foff) & 0x7FFFFFFF
            return self.body_data[foff + 4 : foff + 4 + fsize]

        elif self.case_type == 2:  # case 2: Compressing individually
            fsize = kavutil.get_uint32(self.body_data, foff) & 0x7FFFFFFF
            fdata = self.body_data[foff + 4 : foff + 4 + fsize]
            comp_type = self.check_compression_type(kavutil.get_uint32(fdata, 0))

            if comp_type == self.TYPE_LZMA:
                if HAS_PYLZMA:
                    with contextlib.suppress(TypeError):
                        stream = BytesIO(fdata)
                        obj = pylzma.decompressobj()
                        data = b""

                        while chunk := stream.read(1024):
                            data += obj.decompress(chunk)

                        data += obj.flush()  # Process remaining data
            elif comp_type == self.TYPE_ZLIB:
                if kavutil.get_uint32(self.body_data, foff) & 0x80000000 == 0x80000000:
                    with contextlib.suppress(zlib.error):
                        data = zlib.decompress(fdata, -15)
                else:
                    data = fdata  # TYPE_COPY
        return data

    def close(self):
        if self.mm:
            self.mm.close()
            self.mm = None

        if self.fp:
            self.fp.close()
            self.fp = None

    def get_data(self):
        # NSIS appears to have two types

        # 1. Compressing the entire installation file
        with contextlib.suppress(struct.error):
            head_size = kavutil.get_uint32(self.mm, 0x14)
            comp_size = kavutil.get_uint32(self.mm, 0x18)
            comp_type = self.check_compression_type(kavutil.get_uint32(self.mm, 0x1C))
            if uncmp_data := self.do_decompress(comp_type, 0x1C, comp_size):
                if head_size == kavutil.get_uint32(uncmp_data, 0):
                    self.nsis_header = NSISHeader(uncmp_data[4 : head_size + 4])
                    if self.nsis_header.parse():
                        return uncmp_data[head_size + 4 :], 1

        # case 2: Compressing individually
        with contextlib.suppress(struct.error):
            head_size = kavutil.get_uint32(self.mm, 0x14)
            comp_size = kavutil.get_uint32(self.mm, 0x1C) & 0x7FFFFFFF
            comp_type = self.check_compression_type(kavutil.get_uint32(self.mm, 0x20))
            if uncmp_data := self.do_decompress(comp_type, 0x20, comp_size):
                if head_size == len(uncmp_data):
                    self.nsis_header = NSISHeader(uncmp_data)
                    if self.nsis_header.parse():
                        return self.mm[0x20 + comp_size :], 2

        return None, 0

    def do_decompress(self, comp_type, off, size):
        comp_success = True
        if comp_type == self.TYPE_LZMA:
            if not HAS_PYLZMA:
                comp_success = False
            else:
                try:  # Check if it is compressed
                    stream = BytesIO(self.mm[off : off + size])
                    obj = pylzma.decompressobj()
                    uncmp_data = b""

                    while chunk := stream.read(1024):  # Read 1024 bytes at a time
                        uncmp_data += obj.decompress(chunk)

                    uncmp_data += obj.flush()  # Process remaining data
                except TypeError:
                    comp_success = False
        elif comp_type == self.TYPE_ZLIB:
            try:  # Check if it is compressed
                uncmp_data = zlib.decompress(self.mm[off : off + size], -15)
            except zlib.error:
                comp_success = False
        else:
            uncmp_data = None
            comp_success = False

        if comp_success:
            return uncmp_data

    def check_compression_type(self, data_size):
        """
        Determines the compression type based on data size value

        Args:
            data_size: Size value from NSIS header

        Returns:
            Compression type (LZMA, BZIP2, or ZLIB)
        """
        if data_size & 0x7FFFFFFF == 0x5D:
            return self.TYPE_LZMA
        if data_size & 0xFF == 0x31:
            return self.TYPE_BZIP2
        return self.TYPE_ZLIB

    def __del(self):
        self.close()

        if self.temp_name:
            os.unlink(self.temp_name)


class NSISHeader:
    def __init__(self, data):
        self.mm = data
        self.header_data = None

        self.nh = None
        self.strings_max = 0
        self.ver3 = None
        self.is_unicode = 0  # 1: str, 2:unicode

        # # nsis-2.09-src
        self.ns_skip_code = 252  # 0xfc : to consider next character as a normal character
        self.ns_var_code = 253  # 0xfd : for a variable
        self.ns_shell_code = 254  # 0xfe : for a shell folder path
        self.ns_lang_code = 255  # 0xff : for a langstring

        self.ns_codes_start = 0

        self.files = {}
        self.success = False

    def __set_value(self):
        # Is it a Unicode data set?
        if self.header_data[self.nh.strings : self.nh.strings + 2] == b"\x00\x00":
            self.is_unicode = 2  # Unicode is correct
        else:
            self.is_unicode = 1

        self.strings_max = self.nh.langtables - self.nh.strings  # Maximum string length

        # Check the specific location of the data area in the NSIS header to see if it is Ver3
        off = self.nh.strings + (0x11 * self.is_unicode)
        d = self.header_data[off : off + (14 * self.is_unicode)]

        self.ver3 = self.__binary_unicode_to_str(d) == b"CommonFilesDir"

        if self.ver3:
            self.ns_skip_code = -self.ns_skip_code & 0xFF
            self.ns_var_code = -self.ns_var_code & 0xFF
            self.ns_shell_code = -self.ns_shell_code & 0xFF
            self.ns_lang_code = -self.ns_lang_code & 0xFF
        # end if

        self.ns_codes_start = self.ns_skip_code

        self.__processentries()

    def __binary_str_to_unicode(self, s):
        if self.is_unicode == 2:
            s = s.decode("cp949", errors="replace").encode("utf-16", errors="replace")

        return s

    def __binary_unicode_to_str(self, us):
        if self.is_unicode == 2:
            us = us.decode("utf-16", errors="replace").encode("cp949", errors="replace")

        return us

    def __get_user_var_name(self, n_data):
        if n_data < 0:
            return self.__get_string(n_data)

        static_user_vars = len(NsisVarNames)
        if n_data in range(static_user_vars):
            return f"${NsisVarNames[n_data]}".encode("utf-8")
        else:
            return "$%d".encode("utf-8") % (n_data - static_user_vars)

    def __decode_short(self, off):
        a = int(self.header_data[off])
        b = int(self.header_data[off + 1])

        return ((b & ~0x80) << 7) | (a & ~0x80)

    def __get_string(self, str_off):
        if str_off < 0:
            off = self.nh.langtables + 2 + 4 + -str_off * 4
            str_off = kavutil.get_uint32(self.header_data, off)

        if (str_off * self.is_unicode) in range(self.strings_max):
            return self.__decode_nsis_string(str_off)
        else:
            return ""

    def __decode_nsis_string(self, str_off):
        """
        Decodes a NSIS string from the given offset

        Args:
            str_off: Offset to the string in the header data

        Returns:
            Decoded string with variables expanded

        Handles:
            - Unicode/ASCII string decoding
            - Variable code expansion
            - Special character handling
        """
        str_data = b""
        off = self.nh.strings + (str_off * self.is_unicode)

        char = self.header_data[off : off + (1 * self.is_unicode)]
        off += 1 * self.is_unicode

        while char != b"\x00" * self.is_unicode and len(char) != 0:
            ch = struct.unpack("<H", char)[0] if self.is_unicode == 2 else int.from_bytes(char, byteorder="little")
            # print(off, ch, self.ns_var_code, char)

            if (ch >= self.ns_codes_start) if self.ver3 else (ch < self.ns_codes_start):
                str_data += char

            elif ch == self.ns_var_code:
                n_data = self.__decode_short(off)
                off += 2
                str_data += self.__binary_str_to_unicode(self.__get_user_var_name(n_data))

            char = self.header_data[off : off + (1 * self.is_unicode)]
            off += 1 * self.is_unicode

        str_data = self.__binary_unicode_to_str(str_data)
        return str_data

    def __processentries(self):
        off = self.nh.entries

        for _ in range(self.nh.entries_num):
            # nr = StructNsisRecord()
            # memmove(addressof(nr), self.header_data[off:], sizeof(nr))
            # off += sizeof(nr)  # 28Byte

            val = self.header_data[off : off + 4]

            # if nr.which == 20:  # EW_EXTRACTFILE
            if val == b"\x14\x00\x00\x00":
                nr = StructNsisRecord()
                memmove(addressof(nr), self.header_data[off:], sizeof(nr))

                dt = ""
                with contextlib.suppress(Exception):
                    ft_dec = struct.unpack(">Q", struct.pack(">ll", nr.parm4, nr.parm3))[0]

                    # UnixTimeToFileTime http://support.microsoft.com/kb/167296
                    dt = datetime.datetime.fromtimestamp((ft_dec - 116444736000000000) // 10000000)

                file_name = self.__get_string(nr.parm1).replace(b"\\", b"/")
                self.files[file_name.decode("utf-8")] = nr.parm2, dt, nr.which

            elif val == b"\x3e\x00\x00\x00":
                nr = StructNsisRecord()
                memmove(addressof(nr), self.header_data[off:], sizeof(nr))

                file_name = self.__get_string(nr.parm0).replace(b"\\", b"/")
                # print hex(nr.parm2)
                # print hex(nr.parm3)
                # print hex(nr.parm4)
                # print hex(nr.parm5)

                self.files[file_name.decode("utf-8")] = nr.parm1, "", nr.which

            off += 28

    def parse(self):
        if self.header_data:  # If already analyzed, do not analyze
            return self.success

        self.header_data = self.mm

        self.nh = StructNsisHeader()  # Read the header
        memmove(addressof(self.nh), self.header_data[:], sizeof(self.nh))

        self.__set_value()  # Set important values

        self.success = True
        return self.success

    def namelist(self):
        return self.files.keys()

    def namelist_ex(self):
        fl = []

        for filename in self.files.keys():
            (file_offset, file_time, extract_type) = self.files[filename]
            fl.append((file_offset, filename, file_time, extract_type))

        fl.sort()
        return fl


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """NSIS archive handler plugin.

    This plugin provides functionality for:
    - Detecting NSIS (Nullsoft Scriptable Install System) format
    - Listing files within NSIS installers
    - Extracting files from NSIS installers
    """

    def __init__(self):
        """Initialize the NSIS plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.1",
            title="NSIS Engine",
            kmd_name="nsis",
        )

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["make_arc_type"] = kernel.MASTER_DELETE
        return info

    def __get_handle(self, filename, offset=0):
        """Get or create handle for NSIS file.

        Args:
            filename: Path to NSIS file
            offset: Offset within file

        Returns:
            NSIS object or None
        """
        if filename in self.handle:
            return self.handle.get(filename, None)

        try:
            zfile = NSIS(filename, offset, self.verbose)
            if zfile.parse():
                self.handle[filename] = zfile
                return zfile

        except (IOError, OSError) as e:
            logger.debug("Failed to open NSIS file %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error opening NSIS file %s: %s", filename, e)

        return None

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive.

        Args:
            filename: Path to archive file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        if "ff_nsis" not in fileformat:
            return file_scan_list

        try:
            off = fileformat["ff_nsis"]["Offset"]
            zfile = self.__get_handle(filename, off)
            if zfile is None:
                return file_scan_list

            for name in zfile.namelist():
                # CWE-22: Path traversal prevention
                if k2security.is_safe_archive_member(name):
                    file_scan_list.append(["arc_nsis", name])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive.

        Args:
            arc_engine_id: Engine ID ('arc_nsis')
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        # CWE-22: Path traversal prevention
        if not k2security.is_safe_archive_member(fname_in_arc):
            logger.warning("Unsafe archive member rejected: %s in %s", fname_in_arc, arc_name)
            return None

        if arc_engine_id != "arc_nsis":
            return None

        try:
            zfile = self.__get_handle(arc_name)
            if zfile is None:
                return None

            return zfile.read(fname_in_arc)

        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create an NSIS archive.

        Note: NSIS cannot be re-compressed, so this only marks for deletion.

        Args:
            arc_engine_id: Engine ID ('arc_nsis')
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        if arc_engine_id != "arc_nsis":
            return False

        try:
            # NSIS cannot be re-compressed, so it must be deleted
            return True

        except (IOError, OSError) as e:
            logger.error("Archive creation IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error creating archive %s: %s", arc_name, e)

        return False

    def arcclose(self):
        """Close all open archive handles."""
        for fname in list(self.handle.keys()):
            try:
                zfile = self.handle.get(fname)
                if zfile:
                    zfile.close()
            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
TNEF (Transport Neutral Encapsulation Format) Archive Engine Plugin

This plugin handles TNEF files (winmail.dat) for scanning attachments.
Supports:
- TNEF format detection (signature 0x223E9F78)
- Attachment extraction
- RTF body extraction
- MAPI property parsing
"""

import struct
import os
import re
import logging
from enum import IntEnum
from typing import Optional, Dict, Any, List

from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)

# TNEF signature (little-endian)
TNEF_SIGNATURE = 0x223E9F78


# Attribute levels
class AttrLevel(IntEnum):
    MESSAGE = 0x01
    ATTACHMENT = 0x02


# TNEF attribute IDs
class TnefAttrId(IntEnum):
    # Message level
    TNEF_VERSION = 0x00089006
    OEM_CODEPAGE = 0x00069007
    MESSAGE_CLASS = 0x00078008
    MESSAGE_CLASS2 = 0x00050008
    MAPI_PROPS = 0x00069003
    SUBJECT = 0x00018004
    BODY = 0x0002800C

    # Attachment level
    ATTACH_REND_DATA = 0x00069002
    ATTACH_DATA = 0x0006800F
    ATTACH_TITLE = 0x00018010
    ATTACH_META_FILE = 0x00068011
    ATTACH_CREATE_DATE = 0x00038012
    ATTACH_MODIFY_DATE = 0x00038013
    ATTACHMENT = 0x00069005  # MAPI attachment props


# MAPI property types
class MapiType(IntEnum):
    SHORT = 0x0002
    LONG = 0x0003
    BOOLEAN = 0x000B
    LONGLONG = 0x0014
    DOUBLE = 0x0005
    STRING8 = 0x001E
    UNICODE = 0x001F
    SYSTIME = 0x0040
    BINARY = 0x0102
    CLSID = 0x0048
    MV_FLAG = 0x1000


# MAPI property IDs
class MapiProp(IntEnum):
    RTF_COMPRESSED = 0x1009
    ATTACH_FILENAME = 0x3704
    ATTACH_LONG_FILENAME = 0x3707
    ATTACH_DATA_BIN = 0x3701


# RTF compression magic values
RTF_COMPRESSED_MAGIC = 0x75465A4C  # "LZFu"
RTF_UNCOMPRESSED_MAGIC = 0x414E4D  # "MELA"


class RTFDecompressor:
    """RTF LZFu decompressor."""

    INIT_DICT = (
        b"{\\rtf1\\ansi\\mac\\deff0\\deftab720{\\fonttbl;}"
        b"{\\f0\\fnil \\froman \\fswiss \\fmodern \\fscript "
        b"\\fdecor MS Sans SesijSymbolArialTimes New RomanCourier"
        b"{\\colortbl\\red0\\green0\\blue0\r\n\\par "
        b"\\pard\\plain\\f0\\fs20\\b\\i\\u\\tab\\tx"
    )
    INIT_DICT_SIZE = 207

    @classmethod
    def decompress(cls, data: bytes) -> bytes:
        """Decompress LZFu compressed RTF data."""
        if len(data) < 16:
            return b""

        comp_size = struct.unpack("<I", data[0:4])[0]
        raw_size = struct.unpack("<I", data[4:8])[0]
        magic = struct.unpack("<I", data[8:12])[0]
        crc = struct.unpack("<I", data[12:16])[0]

        if magic == RTF_UNCOMPRESSED_MAGIC:
            return data[16 : 16 + raw_size]

        if magic != RTF_COMPRESSED_MAGIC:
            return b""

        dictionary = bytearray(4096)
        dictionary[: cls.INIT_DICT_SIZE] = cls.INIT_DICT[: cls.INIT_DICT_SIZE]
        write_pos = cls.INIT_DICT_SIZE

        comp_data = data[16:]
        output = bytearray()
        pos = 0

        while pos < len(comp_data) and len(output) < raw_size:
            if pos >= len(comp_data):
                break
            control = comp_data[pos]
            pos += 1

            for i in range(8):
                if pos >= len(comp_data) or len(output) >= raw_size:
                    break

                if control & (1 << i):
                    # Reference - read big-endian (per MS docs)
                    if pos + 1 >= len(comp_data):
                        break
                    ref_hi = comp_data[pos]
                    ref_lo = comp_data[pos + 1]
                    pos += 2

                    offset = (ref_hi << 4) | (ref_lo >> 4)
                    length = (ref_lo & 0x0F) + 2

                    for j in range(length):
                        if len(output) >= raw_size:
                            break
                        byte = dictionary[(offset + j) % 4096]
                        output.append(byte)
                        dictionary[write_pos % 4096] = byte
                        write_pos += 1
                else:
                    byte = comp_data[pos]
                    pos += 1
                    output.append(byte)
                    dictionary[write_pos % 4096] = byte
                    write_pos += 1

        return bytes(output)


class TNEFHandle:
    """TNEF file handle for archive operations."""

    def __init__(self, filename: str):
        """Initialize TNEF handle.

        Args:
            filename: Path to TNEF file
        """
        self.filename = filename
        self.attachments: List[Dict[str, Any]] = []
        self.rtf_body: bytes = b""
        self.codepage = "cp949"

    def open(self) -> bool:
        """Open and parse the TNEF file.

        Returns:
            True if successful, False otherwise
        """
        try:
            with open(self.filename, "rb") as f:
                return self._parse(f)
        except (IOError, OSError) as e:
            logger.debug("Failed to open TNEF file %s: %s", self.filename, e)
            return False
        except Exception as e:
            logger.debug("Failed to parse TNEF file %s: %s", self.filename, e)
            return False

    def close(self):
        """Close the TNEF file handle."""
        self.attachments.clear()
        self.rtf_body = b""

    def _read_bytes(self, fp, size: int) -> bytes:
        """Read exact number of bytes from file."""
        data = fp.read(size)
        if len(data) < size:
            raise ValueError(f"Unexpected EOF: expected {size}, got {len(data)}")
        return data

    def _read_u8(self, fp) -> int:
        return struct.unpack("<B", self._read_bytes(fp, 1))[0]

    def _read_u16(self, fp) -> int:
        return struct.unpack("<H", self._read_bytes(fp, 2))[0]

    def _read_u32(self, fp) -> int:
        return struct.unpack("<I", self._read_bytes(fp, 4))[0]

    def _parse_string(self, data: bytes) -> str:
        """Parse string with null terminator removal."""
        if data.endswith(b"\x00"):
            data = data[:-1]

        for encoding in ["utf-8", self.codepage, "cp1252", "euc-kr", "latin-1"]:
            try:
                return data.decode(encoding)
            except (UnicodeDecodeError, LookupError):
                continue
        return data.decode("latin-1", errors="replace")

    def _parse_mapi_props(self, data: bytes) -> Dict[int, Any]:
        """Parse MAPI properties."""
        props = {}
        pos = 0

        if len(data) < 4:
            return props

        prop_count = struct.unpack("<I", data[0:4])[0]
        pos = 4

        for _ in range(prop_count):
            if pos + 4 > len(data):
                break

            prop_tag = struct.unpack("<I", data[pos : pos + 4])[0]
            pos += 4

            prop_type = prop_tag & 0xFFFF
            prop_id = prop_tag >> 16

            # Named Property handling
            if prop_id >= 0x8000:
                if pos + 16 > len(data):
                    break
                pos += 16  # Skip GUID

                if pos + 4 > len(data):
                    break
                kind = struct.unpack("<I", data[pos : pos + 4])[0]
                pos += 4

                if kind == 0:
                    if pos + 4 > len(data):
                        break
                    pos += 4
                else:
                    if pos + 4 > len(data):
                        break
                    name_len = struct.unpack("<I", data[pos : pos + 4])[0]
                    pos += 4
                    pos += name_len
                    pos = (pos + 3) & ~3

            value = None
            base_type = prop_type & ~MapiType.MV_FLAG

            try:
                if base_type == MapiType.SHORT:
                    if pos + 2 > len(data):
                        break
                    value = struct.unpack("<H", data[pos : pos + 2])[0]
                    pos += 4

                elif base_type == MapiType.LONG:
                    if pos + 4 > len(data):
                        break
                    value = struct.unpack("<I", data[pos : pos + 4])[0]
                    pos += 4

                elif base_type == MapiType.BOOLEAN:
                    if pos + 2 > len(data):
                        break
                    value = struct.unpack("<H", data[pos : pos + 2])[0] != 0
                    pos += 4

                elif base_type == MapiType.LONGLONG:
                    if pos + 8 > len(data):
                        break
                    value = struct.unpack("<Q", data[pos : pos + 8])[0]
                    pos += 8

                elif base_type == MapiType.DOUBLE:
                    if pos + 8 > len(data):
                        break
                    value = struct.unpack("<d", data[pos : pos + 8])[0]
                    pos += 8

                elif base_type == MapiType.SYSTIME:
                    if pos + 8 > len(data):
                        break
                    pos += 8

                elif base_type in (MapiType.STRING8, MapiType.UNICODE):
                    if pos + 4 > len(data):
                        break
                    count = struct.unpack("<I", data[pos : pos + 4])[0]
                    pos += 4

                    values = []
                    for _ in range(count):
                        if pos + 4 > len(data):
                            break
                        str_len = struct.unpack("<I", data[pos : pos + 4])[0]
                        pos += 4
                        if pos + str_len > len(data):
                            break
                        str_data = data[pos : pos + str_len]
                        pos += str_len
                        pos = (pos + 3) & ~3

                        if base_type == MapiType.UNICODE:
                            try:
                                values.append(str_data.decode("utf-16-le").rstrip("\x00"))
                            except Exception:
                                values.append(str_data.decode("latin-1", errors="replace"))
                        else:
                            values.append(self._parse_string(str_data))

                    value = values[0] if len(values) == 1 else values if values else None

                elif base_type == MapiType.BINARY:
                    if pos + 4 > len(data):
                        break
                    count = struct.unpack("<I", data[pos : pos + 4])[0]
                    pos += 4

                    values = []
                    for _ in range(count):
                        if pos + 4 > len(data):
                            break
                        bin_len = struct.unpack("<I", data[pos : pos + 4])[0]
                        pos += 4
                        if pos + bin_len > len(data):
                            break
                        values.append(data[pos : pos + bin_len])
                        pos += bin_len
                        pos = (pos + 3) & ~3

                    value = values[0] if len(values) == 1 else values if values else None

                elif base_type == MapiType.CLSID:
                    if pos + 16 > len(data):
                        break
                    value = data[pos : pos + 16]
                    pos += 16

                else:
                    continue

            except (struct.error, IndexError):
                continue

            if value is not None:
                props[prop_id] = value

        return props

    def _parse_attribute(self, fp) -> Optional[Dict[str, Any]]:
        """Parse a single TNEF attribute."""
        try:
            level = self._read_u8(fp)
            if level not in (AttrLevel.MESSAGE, AttrLevel.ATTACHMENT):
                return None

            attr_id = self._read_u32(fp)
            length = self._read_u32(fp)
            data = self._read_bytes(fp, length)
            checksum = self._read_u16(fp)

            return {"level": level, "attr_id": attr_id, "data": data, "checksum": checksum}
        except (struct.error, ValueError):
            return None

    def _parse(self, fp) -> bool:
        """Parse TNEF file."""
        # Check signature
        signature = self._read_u32(fp)
        if signature != TNEF_SIGNATURE:
            return False

        # Read legacy key
        self._read_u16(fp)

        current_attachment: Optional[Dict[str, Any]] = None

        while True:
            try:
                attr = self._parse_attribute(fp)
                if attr is None:
                    break

                if attr["level"] == AttrLevel.MESSAGE:
                    self._process_message_attribute(attr)
                else:
                    # New attachment start
                    if attr["attr_id"] == TnefAttrId.ATTACH_REND_DATA:
                        if current_attachment and (
                            current_attachment.get("data") or current_attachment.get("filename")
                        ):
                            self.attachments.append(current_attachment)
                        current_attachment = {"filename": "", "long_filename": "", "data": b""}

                    if current_attachment:
                        self._process_attachment_attribute(attr, current_attachment)

            except Exception:
                break

        if current_attachment and (current_attachment.get("data") or current_attachment.get("filename")):
            self.attachments.append(current_attachment)

        return True

    def _process_message_attribute(self, attr: Dict[str, Any]):
        """Process message-level attribute."""
        attr_id = attr["attr_id"]
        data = attr["data"]

        if attr_id == TnefAttrId.OEM_CODEPAGE:
            if len(data) >= 4:
                cp = struct.unpack("<I", data[:4])[0]
                cp_map = {949: "cp949", 932: "cp932", 936: "cp936", 950: "cp950", 1252: "cp1252", 65001: "utf-8"}
                self.codepage = cp_map.get(cp, f"cp{cp}")

        elif attr_id == TnefAttrId.MAPI_PROPS:
            props = self._parse_mapi_props(data)

            if MapiProp.RTF_COMPRESSED in props:
                rtf_data = props[MapiProp.RTF_COMPRESSED]
                self.rtf_body = RTFDecompressor.decompress(rtf_data)

    def _process_attachment_attribute(self, attr: Dict[str, Any], attachment: Dict[str, Any]):
        """Process attachment-level attribute."""
        attr_id = attr["attr_id"]
        data = attr["data"]

        if attr_id == TnefAttrId.ATTACH_TITLE:
            attachment["filename"] = self._parse_string(data)

        elif attr_id == TnefAttrId.ATTACH_DATA:
            attachment["data"] = data

        elif attr_id == TnefAttrId.ATTACHMENT:
            props = self._parse_mapi_props(data)

            if MapiProp.ATTACH_LONG_FILENAME in props:
                attachment["long_filename"] = props[MapiProp.ATTACH_LONG_FILENAME]
            if MapiProp.ATTACH_FILENAME in props and not attachment.get("filename"):
                attachment["filename"] = props[MapiProp.ATTACH_FILENAME]
            if MapiProp.ATTACH_DATA_BIN in props:
                attachment["data"] = props[MapiProp.ATTACH_DATA_BIN]

    def list_files(self) -> List[str]:
        """Get list of extractable files.

        Returns:
            List of filenames (attachments + RTF body if present)
        """
        files = []

        # Add RTF body if present
        if self.rtf_body:
            files.append("message.rtf")

        # Add attachments
        for att in self.attachments:
            filename = att.get("long_filename") or att.get("filename") or "attachment"
            filename = self._sanitize_filename(filename)
            if filename:
                files.append(filename)

        return files

    def extract_file(self, filename: str) -> Optional[bytes]:
        """Extract file data by filename.

        Args:
            filename: File to extract

        Returns:
            File data or None if not found
        """
        # Check RTF body
        if filename == "message.rtf" and self.rtf_body:
            return self.rtf_body

        # Check attachments
        for att in self.attachments:
            att_filename = att.get("long_filename") or att.get("filename") or "attachment"
            att_filename = self._sanitize_filename(att_filename)
            if att_filename == filename:
                return att.get("data")

        return None

    @staticmethod
    def _sanitize_filename(filename: str) -> str:
        """Sanitize filename for safe extraction."""
        filename = re.sub(r'[<>:"/\\|?*\x00-\x1f]', "_", filename)
        filename = filename.strip(". ")
        return filename


def is_valid_tnef(data: bytes) -> bool:
    """Check if data is a valid TNEF file.

    Args:
        data: File data to check

    Returns:
        True if valid TNEF, False otherwise
    """
    if len(data) < 6:
        return False

    try:
        signature = struct.unpack("<I", data[:4])[0]
        return signature == TNEF_SIGNATURE
    except struct.error:
        return False


# -------------------------------------------------------------------------
# class KavMain
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """TNEF archive handler plugin.

    This plugin provides functionality for:
    - Detecting TNEF format (winmail.dat)
    - Listing attachments within TNEF files
    - Extracting attachments and RTF body from TNEF files
    """

    def __init__(self):
        """Initialize the TNEF plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="TNEF Archive Engine",
            kmd_name="tnef",
        )

    def _custom_init(self) -> int:
        """Custom initialization for TNEF plugin.

        Returns:
            0 for success
        """
        return 0

    def _custom_uninit(self) -> int:
        """Custom cleanup for TNEF plugin.

        Returns:
            0 for success
        """
        self.arcclose()
        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["engine_type"] = kernel.ARCHIVE_ENGINE
        return info

    def __get_handle(self, filename: str) -> Optional[TNEFHandle]:
        """Get or create handle for TNEF file.

        Args:
            filename: Path to TNEF file

        Returns:
            TNEFHandle object or None
        """
        if filename in self.handle:
            return self.handle.get(filename)

        tnef_handle = TNEFHandle(filename)
        if tnef_handle.open():
            self.handle[filename] = tnef_handle
            return tnef_handle

        return None

    def format(self, filehandle, filename, filename_ex) -> Optional[Dict[str, Any]]:
        """Analyze and detect TNEF format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to TNEF file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        ret = {}

        try:
            mm = filehandle

            if is_valid_tnef(bytes(mm)):
                ret["ff_tnef"] = "tnef"
                return ret

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return None

    def arclist(self, filename: str, fileformat: Dict[str, Any], password: Optional[str] = None) -> List[List[str]]:
        """List files in the TNEF archive.

        Args:
            filename: Path to TNEF file
            fileformat: Format info from format() method
            password: Not used for TNEF (no encryption support)

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        try:
            if "ff_tnef" in fileformat:
                tnef_handle = self.__get_handle(filename)
                if tnef_handle:
                    for file_name in tnef_handle.list_files():
                        # CWE-22: Path traversal prevention
                        if k2security.is_safe_archive_member(file_name):
                            file_scan_list.append(["arc_tnef", file_name])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id: str, arc_name: str, fname_in_arc: str) -> Optional[bytes]:
        """Extract a file from the TNEF archive.

        Args:
            arc_engine_id: Engine ID ('arc_tnef')
            arc_name: Path to TNEF file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        # CWE-22: Path traversal prevention
        if not k2security.is_safe_archive_member(fname_in_arc):
            logger.debug("Unsafe archive member rejected: %s in %s", fname_in_arc, arc_name)
            return None

        if arc_engine_id != "arc_tnef":
            return None

        try:
            tnef_handle = self.handle.get(arc_name)
            if tnef_handle is None:
                return None

            return tnef_handle.extract_file(fname_in_arc)

        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open TNEF handles."""
        for fname in list(self.handle.keys()):
            try:
                tnef_handle = self.handle[fname]
                if hasattr(tnef_handle, "close"):
                    tnef_handle.close()
            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
Dummy Scan Engine Plugin

This is a test plugin that demonstrates the MalwareDetectorBase interface.
It detects files containing the dummy test pattern.
"""

import os
import logging
from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import MalwareDetectorBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# Malware ID constants for this plugin
# -------------------------------------------------------------------------
class MalwareID:
    """Malware identification constants for dummy plugin."""

    DUMMY_TEST_FILE = 0  # Dummy test file detection


# -------------------------------------------------------------------------
# class KavMain
# -------------------------------------------------------------------------
class KavMain(MalwareDetectorBase):
    """Dummy malware scanner plugin.

    This plugin demonstrates the MalwareDetectorBase interface by detecting
    files that contain a specific test pattern.
    """

    # Dummy pattern to detect
    DUMMY_PATTERN = b"Dummy Engine test file - KICOM Anti-Virus Project"
    DUMMY_VIRUS_NAME = "Dummy-Test-File (not a virus)"

    def __init__(self):
        """Initialize the dummy plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.2",
            title="Dummy Scan Engine",
            kmd_name="dummy",
        )
        self.sig_num = 1
        # Pre-register virus names for listvirus() (before init)
        self.virus_names = [self.DUMMY_VIRUS_NAME]

    def _load_virus_database(self) -> int:
        """Load virus signatures and patterns.

        Returns:
            0 for success, non-zero for failure
        """
        # Register the dummy pattern
        self.virus_patterns[self.DUMMY_VIRUS_NAME] = self.DUMMY_PATTERN
        self.virus_names = [self.DUMMY_VIRUS_NAME]

        if self.verbose:
            logger.info("Dummy plugin: Loaded %d signature(s)", len(self.virus_names))

        return 0

    def scan(self, filehandle, filename, fileformat=None, filename_ex=None):
        """Scan file for malware.

        Overrides base class to maintain compatibility with kernel interface.

        Args:
            filehandle: File data or handle to scan
            filename: Name of the file being scanned
            fileformat: Format information (optional, for compatibility)
            filename_ex: Extended filename (optional, for compatibility)

        Returns:
            Tuple of (threat_found, threat_name, malware_id, scan_state)
        """
        try:
            # Read file data
            if hasattr(filehandle, "read"):
                mm = filehandle.read(len(self.DUMMY_PATTERN))
            elif isinstance(filehandle, bytes):
                mm = filehandle[: len(self.DUMMY_PATTERN)]
            else:
                # Try to open the file directly
                with open(filename, "rb") as fp:
                    mm = fp.read(len(self.DUMMY_PATTERN))

            # Check for dummy pattern
            if mm == self.DUMMY_PATTERN:
                return True, self.DUMMY_VIRUS_NAME, kernel.DISINFECT_DELETE, kernel.INFECTED

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        return False, "", kernel.DISINFECT_NONE, kernel.NOT_FOUND

    def disinfect(self, filename: str, malware_id: int) -> bool:
        """Disinfect (delete) the infected file.

        Args:
            filename: Path to infected file
            malware_id: Malware ID from scan result

        Returns:
            True if successfully deleted, False otherwise
        """
        # Validate path input first
        if not self._validate_path_input(filename, "disinfect_filename"):
            return False

        # Only handle DISINFECT_DELETE
        if malware_id != kernel.DISINFECT_DELETE:
            return False

        try:
            # CWE-73: Safe file deletion
            filename_dir = os.path.dirname(filename) or os.getcwd()
            return k2security.safe_remove_file(filename, filename_dir)
        except (IOError, OSError) as e:
            logger.debug("Disinfect IO error for %s: %s", filename, e)
        except k2security.SecurityError as e:
            logger.warning("Disinfect security error for %s: %s", filename, e)

        return False

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["sig_num"] = self.sig_num
        return info



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
TAR Archive Engine Plugin

This plugin handles TAR archive format for scanning and manipulation.
"""

import contextlib
import os
import re
import tarfile
import logging

from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# class KavMain
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """TAR archive handler plugin.

    This plugin provides functionality for:
    - Detecting TAR archive format
    - Listing files within archives
    - Extracting files from archives
    - Creating/updating archives
    """

    def __init__(self):
        """Initialize the TAR plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.1",
            title="Tar Archive Engine",
            kmd_name="tar",
        )
        self.p_tar_magic = None

    def _custom_init(self) -> int:
        """Custom initialization for TAR plugin.

        Returns:
            0 for success
        """
        self.p_tar_magic = re.compile(rb"[\d\x20\x00]+")
        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["engine_type"] = kernel.ARCHIVE_ENGINE
        info["make_arc_type"] = kernel.MASTER_PACK
        return info

    def __get_handle(self, filename):
        """Get or create handle for TAR file.

        Args:
            filename: Path to TAR file

        Returns:
            TarFile object
        """
        if filename in self.handle:
            return self.handle.get(filename, None)
        else:
            tfile = tarfile.open(filename)
            self.handle[filename] = tfile
            return tfile

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect TAR format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to archive file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        try:
            mm = filehandle
            if p := self.p_tar_magic.match(mm[100:157]):
                if len(p.group()) == 57:
                    ret = {}

                    with contextlib.suppress(tarfile.ReadError):
                        # CWE-404: Use with statement for proper resource cleanup
                        with tarfile.open(filename) as tfile:
                            pass  # Just verify the file can be opened

                        ret["ff_tar"] = "tar"
                        return ret

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return None

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive.

        Args:
            filename: Path to archive file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        try:
            if "ff_tar" in fileformat:
                with contextlib.suppress(tarfile.ReadError):
                    tfile = self.__get_handle(filename)

                    for name in tfile.getnames():
                        # CWE-22: Path traversal prevention
                        if k2security.is_safe_archive_member(name):
                            file_scan_list.append(["arc_tar", name])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive.

        Args:
            arc_engine_id: Engine ID ('arc_tar')
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        # CWE-22: Path traversal prevention
        if not k2security.is_safe_archive_member(fname_in_arc):
            logger.warning("Unsafe archive member rejected: %s in %s", fname_in_arc, arc_name)
            return None

        try:
            if arc_engine_id == "arc_tar":
                tfile = self.__get_handle(arc_name)
                with contextlib.suppress(tarfile.ReadError):
                    f = tfile.extractfile(fname_in_arc)
                    if f is not None:  # extractfile returns None for directories
                        return f.read()

        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create a TAR archive.

        Args:
            arc_engine_id: Engine ID ('arc_tar')
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        if arc_engine_id != "arc_tar":
            return False

        try:
            with tarfile.open(arc_name, "w") as zfile:
                for file_info in file_infos:
                    rname = file_info.get_filename()
                    if not os.path.exists(rname):
                        continue

                    a_name = file_info.get_filename_in_archive()
                    zfile.add(rname, arcname=a_name)
                return True

        except (IOError, OSError) as e:
            logger.error("Archive creation IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error creating archive %s: %s", arc_name, e)

        return False



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
DDE File Format Engine Plugin

This plugin handles DDE (Dynamic Data Exchange) based malware detection.
"""

import contextlib
import logging
import os
import re
import zipfile

from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import MalwareDetectorBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# Returns the data from the decompressed file in the zip file.
# -------------------------------------------------------------------------
def get_zip_data(zip_name, filename):
    data = None

    with contextlib.suppress(zipfile.BadZipfile):
        with zipfile.ZipFile(zip_name) as zfile:
            names = zfile.namelist()

            for name in names:
                if name.lower() == filename:
                    data = zfile.read(name)
                    break

    return data


def InstrSub(obj):
    text = obj.groups()[0]

    off = text.find(b"QUOTE")
    if off != -1:
        t = text[off + 5 :].strip().split(b" ")
        text = b"".join([chr(int(x)) for x in t])

    return text


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(MalwareDetectorBase):
    """DDE malware detector plugin.

    This plugin provides functionality for:
    - Detecting DDE-based exploits in Word documents
    - Detecting CVE-2017-0199 exploits
    """

    def __init__(self):
        """Initialize the DDE plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="DDE Scan Engine",
            kmd_name="dde",
        )
        self.p_dde_text = None
        self.p_instr = None
        self.p_dde = None
        self.p_cmd = None
        self.p_tag = None
        self.p_dde2 = None

    def _load_virus_database(self) -> int:
        """Load virus patterns.

        Returns:
            0 for success
        """
        self.p_dde_text = re.compile(rb'"begin"(.+?)"end"', re.IGNORECASE)
        self.p_instr = re.compile(rb'<w:fldSimple\s+?w:instr=\s*?"(.+?)"\s*>', re.IGNORECASE)
        self.p_dde = re.compile(rb"\bdde(auto)?\b", re.IGNORECASE)
        self.p_cmd = re.compile(rb"\\system32\b(.+)\.exe", re.IGNORECASE)
        self.p_tag = re.compile(rb"\<[\d\D]+?\>")
        self.p_dde2 = re.compile(rb"\x13\s*dde(auto)?\b[^\x00]+", re.IGNORECASE)

        self.virus_names = [
            "Exploit.MSWord.DDE.a",
            "Exploit.MSWord.DDE.b",
            "Exploit.MSWord.CVE-2017-0199",
        ]

        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["sig_num"] = len(self.listvirus())
        return info

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        try:
            if "ff_ooxml" in fileformat:
                if fileformat["ff_ooxml"] == "docx":
                    if data := get_zip_data(filename, "word/document.xml"):
                        if self.__scan_dde_docx(data):
                            return True, "Exploit.MSWord.DDE.a", 0, kernel.INFECTED
                        elif self.__scan_cve_2017_0199_docx(data):
                            return (
                                True,
                                "Exploit.MSWord.CVE-2017-0199",
                                0,
                                kernel.INFECTED,
                            )
            elif filename_ex.lower() == "worddocument":
                data = filehandle
                if self.__scan_dde_doc(data):
                    return True, "Exploit.MSWord.DDE.b", 0, kernel.INFECTED

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        return False, "", -1, kernel.NOT_FOUND

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        try:
            if malware_id == 0:
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)

        return False

    def __scan_dde_docx(self, data):
        """Scans for DDE malware in DOCX files."""
        if not self.p_dde_text:
            return False

        texts = self.p_dde_text.findall(data)
        if len(texts):
            buf = b""
            for text in texts:
                # Remove the front begin Tag
                off = text.find(b">")
                text = text[off + 1 :]

                # Remove the back end Tag
                off = text.rfind(b"<")
                text = text[:off]

                # Process instr
                if self.p_instr:
                    text = self.p_instr.sub(InstrSub, text)

                # Remove all Tags
                if self.p_tag:
                    buf += self.p_tag.sub(b"", text) + b"\n"

            if len(buf) and (self.p_dde and self.p_dde.search(buf) and self.p_cmd and self.p_cmd.search(buf)):
                return True

        return False

    def __scan_dde_doc(self, data):
        """Scans for DDE malware in DOC files."""
        if not self.p_dde2:
            return False

        if s := self.p_dde2.search(data):
            buf = s.group()
            if len(buf) and (self.p_dde and self.p_dde.search(buf) and self.p_cmd and self.p_cmd.search(buf)):
                return True

        return False

    def __scan_cve_2017_0199_docx(self, data):
        """Scans for CVE-2017-0199 malware in DOCX files."""
        return data.find(b'<o:OLEObject Type="Link" ProgID="Word.Document.8"') != -1



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)


import contextlib
import os
import re
import struct
import glob
import marshal
import time
import math
import zlib
import zipfile


# -------------------------------------------------------------------------
# Message printing function
# -------------------------------------------------------------------------
def vprint(header, section=None, msg=None):
    if header:
        print(f"[*] {header}")

    if section:
        new_msg = f"{msg[:22]} ... {msg[-22:]}" if len(msg) > 50 else msg
        print(f"    [-] {section}: {new_msg}")


# -------------------------------------------------------------------------
# Malware pattern instance
# -------------------------------------------------------------------------
handle_pattern_md5 = None  # Malware pattern handle (MD5 hash)
handle_pattern_vdb = None  # Malware pattern handle (VDB)

# -------------------------------------------------------------------------
# Regular expression compilation
# -------------------------------------------------------------------------
p_text = re.compile(r'[\w\s!"#$%&\'()*+,\-./:;<=>?@\[\\\]\^_`{\|}~]')
p_md5_pattern_ext = re.compile(r"\.s(\d\d)$", re.IGNORECASE)


# -------------------------------------------------------------------------
# PatternMD5
# -------------------------------------------------------------------------
class PatternMD5:
    # ---------------------------------------------------------------------
    # __init__(self, rules_paths)
    # Initialize malware pattern
    # input  : rules_paths - Dict with rule paths {"system": "/path", "user": "/path"}
    # ---------------------------------------------------------------------
    def __init__(self, rules_paths):
        self.sig_sizes = {}
        self.sig_p1s = {}
        self.sig_p2s = {}
        self.sig_names = {}
        self.sig_times = {}  # Time information for memory management

        # Validate rules_paths is dict
        if rules_paths is not None and not isinstance(rules_paths, dict):
            raise TypeError(f"rules_paths must be a dict, got {type(rules_paths).__name__}")

        self.rules_paths = rules_paths or {}

        # Number of signatures per malware type: e.g., adware:1, emalware:39
        # This determines which group a size belongs to in the future (size % count) + 1
        self.sig_group_count = {}

        # Collect pattern files from all kicomav rule paths
        pattern_files = []
        for key in ["system", "user"]:
            base = self.rules_paths.get(key)
            if base:
                kicomav_path = os.path.join(base, "kicomav")
                if os.path.isdir(kicomav_path):
                    fl = glob.glob1(kicomav_path, "*.s??")
                    for name in fl:
                        pattern_files.append(os.path.join(kicomav_path, name))

        pattern_files.sort()
        for name in pattern_files:
            if obj := p_md5_pattern_ext.search(name):
                idx = obj.groups()[0]  # ex:01
                sig_key = os.path.split(name)[1].lower().split(".")[0]  # ex:script
                sp = self.__load_sig(name)
                if sp is None:
                    continue

                if len(sp):  # If there is more than one loaded pattern...
                    # Add group count
                    self.sig_group_count[sig_key] = self.sig_group_count.get(sig_key, 0) + 1

                    # Store malware pattern size
                    if sig_key in self.sig_sizes:
                        self.sig_sizes[sig_key].update(dict.fromkeys(sp))
                    else:
                        self.sig_sizes[sig_key] = dict.fromkeys(sp)

    # ---------------------------------------------------------------------
    # match_size(self, sig_key, sig_size)
    # Check if the specified malware pattern exists
    # input  : sig_key  - Specified malware pattern
    #          sig_size - Size
    # return : True or False if the specified size exists in the malware pattern
    # ---------------------------------------------------------------------
    def match_size(self, sig_key, sig_size):
        sig_key = sig_key.lower()  # Convert to lowercase to handle case sensitivity

        return sig_key in self.sig_sizes and sig_size in self.sig_sizes[sig_key]

    # ---------------------------------------------------------------------
    # scan(self, sig_key, sig_size, sig_md5)
    # Scan malware pattern
    # input  : sig_key  - Specified malware pattern
    #          sig_size - Size
    #          sig_md5  - MD5
    # return : Found malware name
    # ---------------------------------------------------------------------
    def scan(self, sig_key, sig_size, sig_md5):
        sig_key = sig_key.lower()  # Convert to lowercase to handle case sensitivity

        if self.match_size(sig_key, sig_size):  # Does the size exist?
            # idxs = self.sig_sizes[sig_key][sig_size]  # Check which file contains the 1st pattern
            idxs = ["%02d" % ((sig_size % self.sig_group_count[sig_key]) + 1)]

            fmd5 = bytes.fromhex(sig_md5)
            sig_p1 = fmd5[:6]  # 1st pattern
            sig_p2 = fmd5[6:]  # 2nd pattern

            for idx in idxs:
                # Compare 1st pattern
                # If 1st pattern is not loaded...
                if self.__load_sig_ex(self.sig_p1s, "i", sig_key, idx) is False:
                    continue

                if sig_p1 in self.sig_p1s[sig_key][idx]:  # Found 1st pattern
                    p2_offs = self.sig_p1s[sig_key][idx][sig_p1]

                    # Compare 2nd pattern
                    # If 2nd pattern is not loaded...
                    if self.__load_sig_ex(self.sig_p2s, "c", sig_key, idx) is False:
                        continue

                    for off in p2_offs:
                        offs = self.sig_p2s[sig_key][idx][off]
                        sig2 = offs[0]  # 2nd pattern
                        if sig2 == sig_p2 and (self.__load_sig_ex(self.sig_names, "n", sig_key, idx) is not False):
                            return self.sig_names[sig_key][idx][offs[1]].decode("utf-8")

        self.__save_mem()  # Reduce memory usage
        return None

    # ---------------------------------------------------------------------
    # __load_sig(self, fname)
    # Load malware pattern
    # input  : fname - Malware pattern file name
    # return : Malware pattern data structure
    # ---------------------------------------------------------------------
    def __load_sig(self, fname):
        try:
            with open(fname, "rb") as fp:
                data = fp.read()
                if data[:4] == b"KAVS":
                    return marshal.loads(zlib.decompress(data[12:]))
        except IOError:
            return None

    # ---------------------------------------------------------------------
    # __load_sig_ex(self, sig_dict, sig_prefix, sig_key, idx)
    # Load malware pattern
    # Also, it can determine whether it is loaded into a specific data structure.
    # input  : sig_dict - Data structure to load malware pattern
    #         sig_prefix - Extension prefix
    #         sig_key    - Malware pattern file name
    #         idx        - Extension number
    # return : Success of loading malware pattern
    # ---------------------------------------------------------------------
    def __load_sig_ex(self, sig_dict, sig_prefix, sig_key, idx):  # (self.sig_names, 'n', 'script', '01')
        if sig_key not in sig_dict:
            sig_dict[sig_key] = {}

        if idx not in sig_dict[sig_key]:
            # Load pattern from kicomav rule paths
            sp = None
            for key in ["system", "user"]:
                base = self.rules_paths.get(key)
                if base:
                    kicomav_path = os.path.join(base, "kicomav")
                    if os.path.isdir(kicomav_path):
                        try:
                            name_fname = os.path.join(kicomav_path, f"{sig_key}.{sig_prefix}{idx}")
                            sp = self.__load_sig(name_fname)
                            if sp is not None:
                                break
                        except IOError:
                            continue

            if sp is None:
                return False

            sig_dict[sig_key][idx] = sp

        # Record current time in sig_time
        if sig_key not in self.sig_times:
            self.sig_times[sig_key] = {}

        if sig_prefix not in self.sig_times[sig_key]:
            self.sig_times[sig_key][sig_prefix] = {}

        self.sig_times[sig_key][sig_prefix][idx] = time.time()

        return True

    # ---------------------------------------------------------------------
    # __save_mem(self)
    # Remove unused malware pattern from memory
    # ---------------------------------------------------------------------
    def __save_mem(self, interval=3 * 60):
        # Is there a pattern to clean up? (3 minutes or more unused pattern)
        n = time.time()
        for sig_key in list(self.sig_times.keys()):
            for sig_prefix in list(self.sig_times[sig_key].keys()):
                for idx in list(self.sig_times[sig_key][sig_prefix].keys()):
                    if n - self.sig_times[sig_key][sig_prefix][idx] > interval:
                        if sig_prefix == "i":  # 1st pattern
                            self.sig_p1s[sig_key].pop(idx)
                        elif sig_prefix == "c":  # 2nd pattern
                            self.sig_p2s[sig_key].pop(idx)
                        elif sig_prefix == "n":  # Malware name pattern
                            self.sig_names[sig_key].pop(idx)

                        self.sig_times[sig_key][sig_prefix].pop(idx)  # Time

    # ---------------------------------------------------------------------
    # get_sig_num(self, sig_key)
    # Get the cumulative number of malware patterns for a given sig_key
    # input  : sig_key - Malware pattern name (ex:script)
    # return : Number of malware patterns
    # ---------------------------------------------------------------------
    def get_sig_num(self, sig_key):
        sig_num = 0

        # Collect count files from all kicomav rule paths
        for key in ["system", "user"]:
            base = self.rules_paths.get(key)
            if base:
                kicomav_path = os.path.join(base, "kicomav")
                if os.path.isdir(kicomav_path):
                    fl = glob.glob1(kicomav_path, f"{sig_key}.c??")
                    for fname in fl:
                        fname = os.path.join(kicomav_path, fname)
                        try:
                            with open(fname, "rb") as fp:
                                buf = fp.read(12)
                                if buf[:4] == b"KAVS":
                                    sig_num += get_uint32(buf, 4)
                        except IOError:
                            continue

        return sig_num

    # ---------------------------------------------------------------------
    # get_sig_vlist(self, sig_key)
    # Get the malware name for a given sig_key
    # input  : sig_key - Malware pattern name (ex:script)
    # return : Malware name
    # ---------------------------------------------------------------------
    def get_sig_vlist(self, sig_key):
        sig_vname = []

        # Collect name files from all kicomav rule paths
        for key in ["system", "user"]:
            base = self.rules_paths.get(key)
            if base:
                kicomav_path = os.path.join(base, "kicomav")
                if os.path.isdir(kicomav_path):
                    fl = glob.glob1(kicomav_path, f"{sig_key}.n??")
                    for fname in fl:
                        fname = os.path.join(kicomav_path, fname)
                        try:
                            sig_vname += self.__load_sig(fname)
                        except IOError:
                            continue

        if not sig_vname:
            return None

        # Convert byte to string
        for i in range(len(sig_vname)):
            sig_vname[i] = sig_vname[i].decode("utf-8")

        return sig_vname


# -------------------------------------------------------------------------
# PatternVDB
# -------------------------------------------------------------------------
class PatternVDB:
    # ---------------------------------------------------------------------
    # __init__(self, rules_paths)
    # Initialize malware pattern
    # input  : rules_paths - Dict with rule paths {"system": "/path", "user": "/path"}
    # ---------------------------------------------------------------------
    def __init__(self, rules_paths):
        self.sig_sizes = {}
        self.sig_p1s = {}
        self.sig_p2s = {}
        self.sig_names = {}
        self.sig_times = {}  # Time information for memory management

        # Validate rules_paths is dict
        if rules_paths is not None and not isinstance(rules_paths, dict):
            raise TypeError(f"rules_paths must be a dict, got {type(rules_paths).__name__}")

        self.rules_paths = rules_paths or {}

        # Collect pattern files from all kicomav rule paths
        pattern_files = []
        for key in ["system", "user"]:
            base = self.rules_paths.get(key)
            if base:
                kicomav_path = os.path.join(base, "kicomav")
                if os.path.isdir(kicomav_path):
                    fl = glob.glob1(kicomav_path, "ve.s??")
                    for name in fl:
                        pattern_files.append(os.path.join(kicomav_path, name))

        pattern_files.sort()
        for name in pattern_files:
            if obj := p_md5_pattern_ext.search(name):
                idx = obj.groups()[0]  # ex:01
                sig_key = os.path.split(name)[1].lower().split(".")[0]  # ex:script
                sp = self.__load_sig(name)
                if sp is None:
                    continue

                if len(sp):  # If more than one loaded pattern...
                    if sig_key not in self.sig_sizes:
                        self.sig_sizes[sig_key] = {}

                    for psize in sp.keys():
                        if psize in self.sig_sizes[sig_key]:
                            self.sig_sizes[sig_key][psize][idx].append(psize)
                        else:
                            self.sig_sizes[sig_key][psize] = {idx: sp[psize]}

    # ---------------------------------------------------------------------
    # match_size(self, sig_key, sig_size)
    # Check if the specified malware pattern exists
    # input  : sig_key  - Specified malware pattern
    #          sig_size - Size
    # return : True or False if the specified size exists in the malware pattern
    # ---------------------------------------------------------------------
    def match_size(self, sig_key, sig_size):
        sig_key = sig_key.lower()  # Convert to lowercase to handle case sensitivity

        if sig_key in self.sig_sizes and sig_size in self.sig_sizes[sig_key].keys():
            return self.sig_sizes[sig_key][sig_size]

        return None

    # ---------------------------------------------------------------------
    # get_cs1(self, ve_id, idx)
    # Read 1st pattern
    # input  : ve_id - ve pattern file
    #          idx   - Internal index
    # return : 1st pattern
    # ---------------------------------------------------------------------
    def get_cs1(self, ve_id, idx):
        sig_key = "ve"

        if self.__load_sig_ex(self.sig_p1s, "i", sig_key, ve_id):
            return self.sig_p1s[sig_key][ve_id][idx]

        return None

    # ---------------------------------------------------------------------
    # get_cs2(self, ve_id, idx)
    # Read 2nd pattern
    # input  : ve_id - ve pattern file
    #          idx   - Internal index
    # return : 2nd pattern
    # ---------------------------------------------------------------------
    def get_cs2(self, ve_id, idx):
        sig_key = "ve"

        if self.__load_sig_ex(self.sig_p2s, "c", sig_key, ve_id):
            return self.sig_p2s[sig_key][ve_id][idx]

        return None

    # ---------------------------------------------------------------------
    # get_vname(self, ve_id, vname_id)
    # Return malware name
    # input  : ve_id    - ve pattern file
    #          vname_id - Internal index
    # return : 1st pattern
    # ---------------------------------------------------------------------
    def get_vname(self, ve_id, vname_id):
        sig_key = "ve"

        if self.__load_sig_ex(self.sig_names, "n", sig_key, ve_id):
            return self.sig_names[sig_key][ve_id][vname_id].decode("utf-8")

        return None

    # ---------------------------------------------------------------------
    # __load_sig(self, fname)
    # Load malware pattern
    # input  : fname - Malware pattern file name
    # return : Malware pattern data structure
    # ---------------------------------------------------------------------
    def __load_sig(self, fname):
        try:
            with open(fname, "rb") as fp:
                data = fp.read()
                if data[:4] == b"KAVS":
                    return marshal.loads(zlib.decompress(data[12:]))
        except IOError:
            return None

    # ---------------------------------------------------------------------
    # __load_sig_ex(self, sig_dict, sig_prefix, sig_key, idx)
    # Load malware pattern
    # Also, it can determine whether it is loaded into a specific data structure.
    # input  : sig_dict - Data structure to load malware pattern
    #         sig_prefix - Extension prefix
    #         sig_key    - Malware pattern file name
    #         idx        - Extension number
    # return : Success of loading malware pattern
    # ---------------------------------------------------------------------
    def __load_sig_ex(self, sig_dict, sig_prefix, sig_key, idx):  # (self.sig_names, 'n', 'script', '01')
        if sig_key not in sig_dict or idx not in sig_dict[sig_key]:
            # Load pattern from kicomav rule paths
            sp = None
            for key in ["system", "user"]:
                base = self.rules_paths.get(key)
                if base:
                    kicomav_path = os.path.join(base, "kicomav")
                    if os.path.isdir(kicomav_path):
                        try:
                            name_fname = os.path.join(kicomav_path, f"{sig_key}.{sig_prefix}{idx}")
                            sp = self.__load_sig(name_fname)
                            if sp is not None:
                                break
                        except IOError:
                            continue

            if sp is None:
                return False

            sig_dict[sig_key] = {idx: sp}

        # Record current time in sig_time
        if sig_key not in self.sig_times:
            self.sig_times[sig_key] = {}

        if sig_prefix not in self.sig_times[sig_key]:
            self.sig_times[sig_key][sig_prefix] = {}

        self.sig_times[sig_key][sig_prefix][idx] = time.time()

        return True

    # ---------------------------------------------------------------------
    # __save_mem(self)
    # Remove unused malware pattern from memory
    # ---------------------------------------------------------------------
    def __save_mem(self, interval=3 * 60):
        # Is there a pattern to clean up? (3 minutes or more unused pattern)
        n = time.time()
        for sig_key in list(self.sig_times.keys()):
            for sig_prefix in list(self.sig_times[sig_key].keys()):
                for idx in list(self.sig_times[sig_key][sig_prefix].keys()):
                    if n - self.sig_times[sig_key][sig_prefix][idx] > interval:
                        if sig_prefix == "i":  # 1st pattern
                            self.sig_p1s[sig_key].pop(idx)
                        elif sig_prefix == "c":  # 2nd pattern
                            self.sig_p2s[sig_key].pop(idx)
                        elif sig_prefix == "n":  # Malware name pattern
                            self.sig_names[sig_key].pop(idx)

                        self.sig_times[sig_key][sig_prefix].pop(idx)  # Time

    # ---------------------------------------------------------------------
    # get_sig_num(self, sig_key)
    # Get the cumulative number of malware patterns for a given sig_key
    # input  : sig_key - Malware pattern name (ex:script)
    # return : Number of malware patterns
    # ---------------------------------------------------------------------
    def get_sig_num(self, sig_key):
        sig_num = 0

        # Collect count files from all kicomav rule paths
        for key in ["system", "user"]:
            base = self.rules_paths.get(key)
            if base:
                kicomav_path = os.path.join(base, "kicomav")
                if os.path.isdir(kicomav_path):
                    fl = glob.glob1(kicomav_path, f"{sig_key}.c??")
                    for fname in fl:
                        fname = os.path.join(kicomav_path, fname)
                        try:
                            with open(fname, "rb") as fp:
                                buf = fp.read(12)
                                if buf[:4] == b"KAVS":
                                    sig_num += get_uint32(buf, 4)
                        except IOError:
                            continue

        return sig_num

    # ---------------------------------------------------------------------
    # get_sig_vlist(self, sig_key)
    # Get the malware name for a given sig_key
    # input  : sig_key - Malware pattern name (ex:script)
    # return : Malware name
    # ---------------------------------------------------------------------
    def get_sig_vlist(self, sig_key):
        sig_vname = []

        # Collect name files from all kicomav rule paths
        for key in ["system", "user"]:
            base = self.rules_paths.get(key)
            if base:
                kicomav_path = os.path.join(base, "kicomav")
                if os.path.isdir(kicomav_path):
                    fl = glob.glob1(kicomav_path, f"{sig_key}.n??")
                    for fname in fl:
                        fname = os.path.join(kicomav_path, fname)
                        try:
                            sig_vname += self.__load_sig(fname)
                        except IOError:
                            continue

        if not sig_vname:
            return None

        # Convert byte to string
        for i in range(len(sig_vname)):
            sig_vname[i] = sig_vname[i].decode("utf-8")

        return sig_vname


# -------------------------------------------------------------------------
# AhoCorasick class
# Source : https://gist.github.com/atdt/875e0dba6a15e3fa6018
# -------------------------------------------------------------------------
FAIL = -1


class AhoCorasick:
    def __init__(self):
        self.transitions = {}
        self.outputs = {}
        self.fails = {}

    def make_tree(self, keywords):
        new_state = 0

        for keyword in keywords:
            state = 0

            for j, char in enumerate(keyword):
                res = self.transitions.get((state, char), FAIL)
                if res == FAIL:
                    break
                state = res

            for char in keyword[j:]:
                new_state += 1
                self.transitions[(state, char)] = new_state
                state = new_state

            self.outputs[state] = [keyword]

        queue = []
        for (from_state, char), to_state in self.transitions.items():
            if from_state == 0 and to_state != 0:
                queue.append(to_state)
                self.fails[to_state] = 0

        while queue:
            r = queue.pop(0)
            for (from_state, char), to_state in self.transitions.items():
                if from_state == r:
                    queue.append(to_state)
                    state = self.fails[from_state]

                    while True:
                        res = self.transitions.get((state, char), state and FAIL)
                        if res != FAIL:
                            break
                        state = self.fails[state]

                    failure = self.transitions.get((state, char), state and FAIL)
                    self.fails[to_state] = failure
                    self.outputs.setdefault(to_state, []).extend(self.outputs.get(failure, []))

    def search(self, string):
        state = 0
        results = []
        for i, char in enumerate(string):
            while True:
                res = self.transitions.get((state, char), state and FAIL)
                if res != FAIL:
                    state = res
                    break
                state = self.fails[state]

            for match in self.outputs.get(state, ()):
                pos = i - len(match) + 1
                results.append((pos, match))

        return results


# -------------------------------------------------------------------------
# HexDump
# Display Hex dump for a specified area in a given file
# input  : fname : File name
#          start : Start position of the area to dump
#          size  : Size to dump
#          width : Number of characters per line
# -------------------------------------------------------------------------
class HexDump:
    def File(self, fname, start, size=0x200, width=16):
        with open(fname, "rb") as fp:
            fp.seek(start)
            buf = fp.read(size)
            self.Buffer(buf, start, size, width)

    # -------------------------------------------------------------------------
    # Buffer
    # Display Hex dump for a given buffer
    # input  : fbuf   : Buffer
    #          start : Start position of the area to dump
    #          size  : Size to dump
    #          width : Number of characters per line
    # -------------------------------------------------------------------------
    def Buffer(self, buf, start, size=0x200, width=16):
        # If the buffer is larger than the given size, adjust the argument
        if len(buf) < size:
            size = len(buf)
        row = start % width  # Column
        col = start // width  # Row
        # [row ... width*col]
        # [width*col ... width * (col+1)]
        r_size = 0
        line_start = row + (col * width)
        while True:
            line = buf[line_start : width * (col + 1)]

            if len(line) == 0:
                break
            if r_size + len(line) >= size:
                line = line[: size - r_size]
                r_size = size - len(line)
            # Address value
            output = "%08X : " % ((line_start // width) * width)
            # Hex value
            output += row * "   " + "".join("%02x " % c for c in line)
            output += "  " + (width - (row + len(line))) * "   "
            # Character value
            output += row * " "
            output += "".join([".", chr(c)][self.IsPrint(c)] for c in line)
            print(output)
            line_start = width * (col + 1)
            col += 1
            row = 0
            r_size += len(line)
            if r_size == size:
                break

    # -------------------------------------------------------------------------
    # IsPrint
    # Check if the given character is printable
    # input  : char  : Character
    # return : True  : Printable character
    #          False : Non-printable character
    # -------------------------------------------------------------------------
    def IsPrint(self, c):
        return c >= 0x20 and c < 0x80


# -------------------------------------------------------------------------
# is_textfile(buf)
# Determine if the given buffer is a text file
# input  : buf - Buffer
# return : True if it is a text file, False otherwise
# -------------------------------------------------------------------------
def is_textfile(buf):
    n_buf = len(buf)

    if n_buf == 0:
        return False

    buf = buf.decode("latin-1")

    n_text = len(p_text.findall(buf))

    return n_text / float(n_buf) > 0.8


# -------------------------------------------------------------------------
# get_uint16(buf, off):
# Read the value from the given buffer based on the offset as uint16
# input  : buf - Buffer
#          off - Offset
# return : uint16 converted value
# -------------------------------------------------------------------------
def get_uint16(buf, off, endian="<"):
    return struct.unpack(f"{endian}H", buf[off : off + 2])[0]


# -------------------------------------------------------------------------
# get_uint32(buf, off):
# Read the value from the given buffer based on the offset as uint32
# input  : buf - Buffer
#          off - Offset
# return : uint32 converted value
# -------------------------------------------------------------------------
def get_uint32(buf, off, endian="<"):
    return struct.unpack(f"{endian}L", buf[off : off + 4])[0]


# -------------------------------------------------------------------------
# get_uint64(buf, off):
# Read the value from the given buffer based on the offset as uint64
# input  : buf - Buffer
#          off - Offset
# return : uint64 converted value
# -------------------------------------------------------------------------
def get_uint64(buf, off, endian="<"):
    return struct.unpack(f"{endian}Q", buf[off : off + 8])[0]


# -------------------------------------------------------------------------
# normal_vname(vname):
# Process special characters in the given malware name
# input  : vname - Malware name
#          platform - Win32, MSIL, etc.
# return : New malware name
# -------------------------------------------------------------------------
def normal_vname(vname, platform=None):
    # vname = vname.replace('<n>', 'not-a-virus:')
    vname = vname.replace("<n>", "")

    if platform:
        vname = vname.replace("<p>", platform)

    return vname


# -------------------------------------------------------------------------
# uniq_string()
# Return a unique string in numeric format
# -------------------------------------------------------------------------
def uniq_string():
    return str(int(time.time() * 1000))


# ----------------------------------------------------------------------------
# make_zip(arc_name, file_infos)
# Logic to disinfect malware in a compressed file and re-compress it
# Actual compression logic is not implemented in non-compression engines (e.g., ALZ, EGG, etc.)
# The file is compressed using ZIP, and the extension is retained
# ----------------------------------------------------------------------------
def make_zip(arc_name, file_infos):
    with open(arc_name, "rb") as fp:
        if fp.read(2) != b"PK":
            return

    # CWE-404: Use with statement for proper resource cleanup
    with zipfile.ZipFile(arc_name, "w") as zfile:
        for file_info in file_infos:
            rname = file_info.get_filename()
            with contextlib.suppress(IOError):
                with open(rname, "rb") as fp:
                    buf = fp.read()
                    a_name = file_info.get_filename_in_archive()
                    zfile.writestr(a_name, buf, compress_type=zipfile.ZIP_DEFLATED)


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain:
    # ---------------------------------------------------------------------
    # init(self, rules_paths)
    # Initialize the plug-in engine
    # input  : rules_paths - Dict with rule paths {"system": "/path", "user": "/path"}
    #          verbose     - Debug mode (True or False)
    # return : 0 - success, non-zero - failure
    # ---------------------------------------------------------------------
    def init(self, rules_paths=None, verbose=False):  # Initialize the plug-in engine
        # Validate rules_paths is dict
        if rules_paths is not None and not isinstance(rules_paths, dict):
            raise TypeError(f"rules_paths must be a dict, got {type(rules_paths).__name__}")

        # Initialize malware pattern
        global handle_pattern_md5
        global handle_pattern_vdb

        handle_pattern_md5 = PatternMD5(rules_paths)
        handle_pattern_vdb = PatternVDB(rules_paths)

        return 0  # Initialize the plug-in engine successfully

    # ---------------------------------------------------------------------
    # uninit(self)
    # Terminate the plug-in engine
    # return : 0 - success, Nonzero - fail
    # ---------------------------------------------------------------------
    def uninit(self):  # Terminate the plug-in engine
        return 0  # Terminate the plug-in engine successfully

    # ---------------------------------------------------------------------
    # getinfo(self)
    # Get the main information of the plug-in engine (author, version, ...)
    # return : Plug-in engine information
    # ---------------------------------------------------------------------
    def getinfo(self):  # Get the main information of the plug-in engine
        return {
            "author": "Kei Choi",
            "version": "1.1",
            "title": "KicomAV Utility Library",
            "kmd_name": "kavutil",
        }



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
Attach Engine Plugin

This plugin handles attached data extraction from files.
"""

import contextlib
import os
import logging

from kicomav.plugins import kernel
from kicomav.plugins import kavutil
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# class KavMain
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """Attach data handler plugin.

    This plugin provides functionality for:
    - Extracting attached data from files
    - Creating files with attached data removed
    """

    def __init__(self):
        """Initialize the Attach plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.1",
            title="Attach Engine",
            kmd_name="attach",
        )

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["make_arc_type"] = kernel.MASTER_PACK
        return info

    def arclist(self, filename, fileformat, password=None):
        """List attached data in the file.

        Args:
            filename: Path to file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        if "ff_attach" in fileformat:
            pos = fileformat["ff_attach"]["Attached_Pos"]
            size = fileformat["ff_attach"]["Attached_Size"]
            file_scan_list.append([f"arc_attach:{pos}:{size}", "Attached"])

            if self.verbose:
                self._print_attach_debug_info(filename, pos, size)

        return file_scan_list

    def _print_attach_debug_info(self, filename, pos, size):
        """Print debug information about attached data."""
        print("-" * 79)
        kavutil.vprint("Engine")
        kavutil.vprint(None, "Engine", "attach")
        kavutil.vprint(None, "File name", os.path.split(filename)[-1])
        kavutil.vprint(None, "Attach Point", "0x%08X" % pos)
        kavutil.vprint(None, "Attach Size", "0x%08X" % size)

        try:
            with open(filename, "rb") as fp:
                fp.seek(pos)
                buf = fp.read(0x80)
                print()
                kavutil.vprint("Attach Point (Raw)")
                print()
                kavutil.HexDump().Buffer(buf, 0, 0x80)
        except (IOError, OSError) as e:
            logger.debug("Error reading attach point: %s", e)

        print()

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract attached data from the file.

        Args:
            arc_engine_id: Engine ID (arc_attach:pos:size)
            arc_name: Path to file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        try:
            if arc_engine_id.startswith("arc_attach:"):
                t = arc_engine_id.split(":")
                pos = int(t[1])
                size = int(t[2])

                with open(arc_name, "rb") as fp:
                    fp.seek(pos)
                    return fp.read(size)

        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create file with attached data removed or replaced.

        Args:
            arc_engine_id: Engine ID (arc_attach:pos:size)
            arc_name: Path to file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        if not arc_engine_id.startswith("arc_attach:"):
            return False

        try:
            t = arc_engine_id.split(":")
            pos = int(t[1])

            file_info = file_infos[0]
            rname = file_info.get_filename()

            # Read original file
            with open(arc_name, "rb") as fp:
                t_buf = fp.read()

            # Build new content
            if os.path.exists(rname):
                with open(rname, "rb") as fp:
                    buf = fp.read()
                data = t_buf[:pos] + buf
            else:
                data = t_buf[:pos]

            # Write clean file
            with open(arc_name, "wb") as wp:
                wp.write(data)

            return True

        except (IOError, OSError) as e:
            logger.error("Archive creation IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error creating archive %s: %s", arc_name, e)

        return False



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)
# Security utilities for KicomAV Engine

"""
KicomAV Security Utilities

his module provides the following security features:
- Path Traversal prevention (CWE-22)
- Zip Slip prevention
- File name validation
- Safe file deletion
- URL validation (SSRF prevention)
- File hash calculation
"""

import os
import hashlib
import pathlib
import ipaddress
import urllib.parse
import urllib.request
from typing import Optional, List, Callable


# -------------------------------------------------------------------------
# SecurityError Exception
# -------------------------------------------------------------------------
class SecurityError(Exception):
    """Security exception class"""

    pass


# -------------------------------------------------------------------------
# Path Validation Functions
# -------------------------------------------------------------------------
def validate_path(file_path: str, base_dir: str, allow_symlinks: bool = False) -> str:
    """
    Path validation - Path Traversal prevention (CWE-22)

    Args:
        file_path: Path to validate
        base_dir: Allowed base directory
        allow_symlinks: Allow symbolic links

    Returns:
        Normalized safe path

    Raises:
        SecurityError: Path validation failed
    """
    # 1. Convert to absolute path and normalize
    file_path = os.path.abspath(os.path.normpath(file_path))
    base_dir = os.path.abspath(os.path.normpath(base_dir))

    # 2. Verify if the path is inside base_dir
    # Use os.sep to support both Windows and Unix
    if not (file_path.startswith(base_dir + os.sep) or file_path == base_dir):
        raise SecurityError(f"Path outside base directory: {file_path}")

    # 3. Verify ../ pattern (check after normalization)
    path_parts = pathlib.Path(file_path).parts
    if ".." in path_parts:
        raise SecurityError(f"Parent directory reference in path: {file_path}")

    # 4. Verify symbolic links
    if not allow_symlinks and os.path.islink(file_path):
        real_path = os.path.realpath(file_path)
        if not (real_path.startswith(base_dir + os.sep) or real_path == base_dir):
            raise SecurityError(f"Symlink points outside base directory: {file_path}")

    return file_path


def safe_extract_member(archive_member_name: str, extract_base_dir: str) -> str:
    """
    Safe extraction of archive member - Zip Slip prevention (CWE-22)

    Args:
        archive_member_name: Name of the archive member
        extract_base_dir: Base directory for extraction

    Returns:
        Safe extraction path

    Raises:
        SecurityError: Path validation failed
    """
    # 1. Verify null byte
    if "\0" in archive_member_name:
        raise SecurityError(f"Null byte in filename: {archive_member_name}")

    # 2. Block absolute path
    if os.path.isabs(archive_member_name):
        raise SecurityError(f"Absolute path in archive: {archive_member_name}")

    # 3. Normalize path
    extract_path = os.path.normpath(os.path.join(extract_base_dir, archive_member_name))

    # 4. Verify path traversal
    return validate_path(extract_path, extract_base_dir, allow_symlinks=False)


# -------------------------------------------------------------------------
# Filename Validation Functions
# -------------------------------------------------------------------------
def safe_filename(filename: str, max_length: int = 255) -> bool:
    """
    Verify filename safety

    Args:
        filename: Filename to verify
        max_length: Maximum length

    Returns:
        Verification result (True)

    Raises:
        SecurityError: Verification failed
    """
    # 1. Verify null byte
    if "\0" in filename:
        raise SecurityError("Null byte in filename")

    # 2. Verify path separator
    if "/" in filename or "\\" in filename:
        raise SecurityError("Path separator in filename")

    # 3. Verify length
    if len(filename) > max_length:
        raise SecurityError(f"Filename too long: {len(filename)} > {max_length}")

    # 4. Verify dangerous characters (Windows reserved characters)
    dangerous_chars = ["<", ">", ":", '"', "|", "?", "*"]
    if any(c in filename for c in dangerous_chars):
        raise SecurityError(f"Dangerous character in filename: {filename}")

    # 5. Verify reserved names (Windows)
    reserved_names = [
        "CON",
        "PRN",
        "AUX",
        "NUL",
        "COM1",
        "COM2",
        "COM3",
        "COM4",
        "COM5",
        "COM6",
        "COM7",
        "COM8",
        "COM9",
        "LPT1",
        "LPT2",
        "LPT3",
        "LPT4",
        "LPT5",
        "LPT6",
        "LPT7",
        "LPT8",
        "LPT9",
    ]
    # Compare with the name without extension
    name_without_ext = os.path.splitext(filename)[0].upper()
    if name_without_ext in reserved_names:
        raise SecurityError(f"Reserved filename: {filename}")

    # 6. Verify empty filename
    if not filename or filename.strip() == "":
        raise SecurityError("Empty filename")

    # 7. Verify if the filename starts or ends with a period (Windows issue)
    if filename.endswith(".") or filename.endswith(" "):
        raise SecurityError(f"Filename ends with invalid character: {filename}")

    return True


# -------------------------------------------------------------------------
# File Operations
# -------------------------------------------------------------------------
def safe_remove_file(filepath: str, allowed_dir: str) -> bool:
    """
    Safe file deletion (CWE-73)

    Args:
        filepath: Path to delete
        allowed_dir: Allowed directory

    Returns:
        Deletion success status

    Raises:
        SecurityError: Verification failed
    """
    # 1. Verify path
    filepath = validate_path(filepath, allowed_dir, allow_symlinks=False)

    # 2. Verify file existence
    if not os.path.exists(filepath):
        return False

    # 3. Verify if it is a file (prevent directory deletion)
    if not os.path.isfile(filepath):
        raise SecurityError(f"Not a file: {filepath}")

    # 4. Verify if it is a symbolic link
    if os.path.islink(filepath):
        raise SecurityError(f"Cannot delete symlink: {filepath}")

    # 5. Safely delete
    os.remove(filepath)
    return True


def hash_file_sha256(filepath: str) -> str:
    """
    Calculate SHA256 hash of a file

    Args:
        filepath: Path to calculate hash

    Returns:
        SHA256 hash (hex string)
    """
    sha256_hash = hashlib.sha256()
    with open(filepath, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


# -------------------------------------------------------------------------
# URL Validation Functions
# -------------------------------------------------------------------------
def validate_url(
    url: str, allowed_schemes: Optional[List[str]] = None, allowed_domains: Optional[List[str]] = None
) -> bool:
    """
    URL validation - SSRF prevention (CWE-918)

    Args:
        url: URL to validate
        allowed_schemes: Allowed schemes list (default: ['https'])
        allowed_domains: Allowed domains list

    Returns:
        Verification result (True)

    Raises:
        SecurityError: Verification failed
    """
    if allowed_schemes is None:
        allowed_schemes = ["https"]

    # 1. Parse URL
    try:
        parsed = urllib.parse.urlparse(url)
    except Exception as e:
        raise SecurityError(f"Invalid URL: {e}")

    # 2. Verify scheme
    if parsed.scheme not in allowed_schemes:
        raise SecurityError(f"Disallowed URL scheme: {parsed.scheme}")

    # 3. Extract hostname
    hostname = parsed.hostname
    if not hostname:
        raise SecurityError("No hostname in URL")

    # 4. Verify IP address (block Private IP)
    try:
        ip = ipaddress.ip_address(hostname)
        if ip.is_private:
            raise SecurityError(f"Private IP address: {ip}")
        if ip.is_loopback:
            raise SecurityError(f"Loopback IP address: {ip}")
        if ip.is_link_local:
            raise SecurityError(f"Link-local IP address: {ip}")
        if ip.is_reserved:
            raise SecurityError(f"Reserved IP address: {ip}")
        # Block cloud metadata server (169.254.169.254)
        if str(ip) == "169.254.169.254":
            raise SecurityError(f"Cloud metadata IP address blocked: {ip}")
    except ValueError:
        # If it is a domain name, continue
        pass

    # 5. Verify domain whitelist
    if allowed_domains is not None:
        domain_allowed = False
        for domain in allowed_domains:
            if hostname == domain or hostname.endswith("." + domain):
                domain_allowed = True
                break
        if not domain_allowed:
            raise SecurityError(f"Domain not in whitelist: {hostname}")

    # 6. Block dangerous hostname
    dangerous_hosts = ["localhost", "127.0.0.1", "0.0.0.0", "::1"]
    if hostname.lower() in dangerous_hosts:
        raise SecurityError(f"Dangerous hostname: {hostname}")

    return True


# -------------------------------------------------------------------------
# Archive Security Functions
# -------------------------------------------------------------------------
def is_safe_archive_member(member_name: str) -> bool:
    """
    Verify the safety of the archive member name quickly

    Args:
        member_name: Name of the archive member

    Returns:
        Safety status (True/False)
    """
    # Verify null byte
    if "\0" in member_name:
        return False

    # Verify absolute path
    if os.path.isabs(member_name):
        return False

    # Verify path traversal
    if ".." in pathlib.Path(member_name).parts:
        return False

    return True


def get_safe_extract_path(archive_member_name: str, extract_base_dir: str) -> Optional[str]:
    """
    Return a safe extraction path, or return None if it is not safe

    Args:
        archive_member_name: Name of the archive member
        extract_base_dir: Base directory for extraction

    Returns:
        Safe extraction path or None
    """
    try:
        return safe_extract_member(archive_member_name, extract_base_dir)
    except SecurityError:
        return None


# -------------------------------------------------------------------------
# Secure Download Functions (CWE-434, CWE-918)
# -------------------------------------------------------------------------
def safe_download_file(
    url: str,
    dest_path: str,
    base_dir: str,
    allowed_schemes: Optional[List[str]] = None,
    allowed_domains: Optional[List[str]] = None,
    fnhook: Optional[Callable] = None,
) -> str:
    """
    Secure file download (CWE-434, CWE-918 prevention)

    Validates URL for SSRF prevention and destination path for path traversal.

    Args:
        url: URL to download from
        dest_path: Destination file path
        base_dir: Allowed base directory for downloads
        allowed_schemes: Allowed URL schemes (default: ["https", "http"])
        allowed_domains: Allowed domains list (None = allow all public domains)
        fnhook: Progress callback function

    Returns:
        Safe download path

    Raises:
        SecurityError: URL or path validation failed
    """
    # Default to allow both http and https for updates
    if allowed_schemes is None:
        allowed_schemes = ["https", "http"]

    # 1. Validate URL (SSRF prevention - CWE-918)
    validate_url(url, allowed_schemes=allowed_schemes, allowed_domains=allowed_domains)

    # 2. Validate destination path (Path Traversal prevention - CWE-22)
    safe_dest = validate_path(dest_path, base_dir)

    # 3. Download file securely
    try:
        if fnhook is not None:
            urllib.request.urlretrieve(url, safe_dest, fnhook)
        else:
            urllib.request.urlretrieve(url, safe_dest)
    except Exception as e:
        raise SecurityError(f"failed")

    return safe_dest


def validate_download_filename(filename: str) -> str:
    """
    Validate download filename for dangerous characters and extensions

    Args:
        filename: Filename to validate

    Returns:
        Validated filename

    Raises:
        SecurityError: Invalid filename
    """
    # Check for null byte
    if "\0" in filename:
        raise SecurityError("Null byte in filename")

    # Check for path separators
    if "/" in filename or "\\" in filename:
        raise SecurityError("Path separator in filename")

    # Check for path traversal
    if ".." in filename:
        raise SecurityError("Path traversal in filename")

    # Check for empty filename
    if not filename or filename.strip() == "":
        raise SecurityError("Empty filename")

    # Check for reserved device names on Windows
    reserved_names = [
        "CON",
        "PRN",
        "AUX",
        "NUL",
        "COM1",
        "COM2",
        "COM3",
        "COM4",
        "COM5",
        "COM6",
        "COM7",
        "COM8",
        "COM9",
        "LPT1",
        "LPT2",
        "LPT3",
        "LPT4",
        "LPT5",
        "LPT6",
        "LPT7",
        "LPT8",
        "LPT9",
    ]
    name_without_ext = os.path.splitext(filename)[0].upper()
    if name_without_ext in reserved_names:
        raise SecurityError(f"Reserved filename: {filename}")

    return filename



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
RTF File Format Engine Plugin

This plugin handles RTF format for scanning, malware detection, and extraction.
"""

import contextlib
import logging
import os
import re
from pathlib import Path

from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# Functions related to extracting objdata
# -------------------------------------------------------------------------
HEX_RE = re.compile(rb"[0-9A-Fa-f]{2}")


# -------------------------------------------------------------------------
# RtfFile class
# -------------------------------------------------------------------------
class RtfFile:
    def __init__(self, filename, verbose=False):
        self.verbose = verbose  # Debugging mode
        self.filename = filename

        self.p = re.compile(rb"[A-Fa-f0-9]+")

        self.num_objdata = 0  # Number of objdata inserted into RTF
        self.objdata = {}  # objdata
        self.payloads = []  # Extracted payloads
        self.parse()

    def parse(self):
        self.payloads = self.extract_objdata_payloads()
        for idx, payload in enumerate(self.payloads):
            fname = f"Objdata #{idx + 1}"
            self.objdata[fname] = payload

        self.num_objdata = len(self.payloads)

    def extract_objdata_payloads(self) -> list[bytes]:
        payloads: list[bytes] = []
        rtf_bytes = Path(self.filename).read_bytes()

        for m in re.finditer(rb"\\objdata\b", rtf_bytes):
            pos = m.start()
            group_start = rtf_bytes.rfind(b"{", 0, pos)
            if group_start == -1:
                continue

            try:
                group_end = self._find_matching_brace(rtf_bytes, group_start)
            except ValueError:
                continue

            group = rtf_bytes[group_start : group_end + 1]

            # Parse only after \objdata in the group
            objdata_idx = group.find(b"\\objdata")
            if objdata_idx == -1:
                continue
            i = objdata_idx + len(b"\\objdata")
            out = bytearray()

            while i < len(group):
                # Process \binN
                if group[i : i + 4] == b"\\bin":
                    j = i + 4
                    # Read the number (decimal)
                    num_start = j
                    while j < len(group) and group[j : j + 1].isdigit():
                        j += 1
                    if j == num_start:
                        i += 4
                        continue

                    nbytes = int(group[num_start:j].decode("ascii", errors="ignore") or "0")
                    j = self._skip_whitespace(group, j)
                    # Copy the next nbytes as is
                    out += group[j : j + nbytes]
                    i = j + nbytes
                    continue

                # End of group
                if group[i : i + 1] == b"}":
                    break

                # Collect only 2-hex tokens (ignore whitespace/line breaks/other control characters)
                # If the current position is a hex 2-digit, select it
                if i + 2 <= len(group) and HEX_RE.fullmatch(group[i : i + 2]):
                    out.append(int(group[i : i + 2], 16))
                    i += 2
                    continue

                i += 1

            if out:
                payloads.append(bytes(out))

        return payloads

    def _find_matching_brace(self, data: bytes, start: int) -> int:
        """
        start is the position of '{'. Return the position of the matching '}'.
        In RTF, \{ \} are ignored (simple processing).
        """
        depth = 0
        i = start
        n = len(data)

        while i < n:
            ch = data[i : i + 1]

            if ch == b"{":
                depth += 1
            elif ch == b"}":
                depth -= 1
                if depth == 0:
                    return i

            i += 1

        raise ValueError("No matching '}' found for objdata group.")

    def _skip_whitespace(self, data: bytes, i: int) -> int:
        n = len(data)
        while i < n and data[i] in b" \t\r\n":
            i += 1
        return i

    def close(self):
        pass

    def namelist(self):
        names = []

        if len(self.objdata):
            names = list(self.objdata.keys())
            names.sort()

        return names

    def read(self, fname):
        return self.objdata.get(fname, None)


# -------------------------------------------------------------------------
# RtfPackage class
# -------------------------------------------------------------------------
class RtfPackage:
    def __init__(self, filename, verbose=False):
        self.verbose = verbose  # Debugging mode
        self.filename = filename

        self.objdata = {}  # objdata
        self.payloads = []  # Extracted payloads
        self.parse()

    def parse(self):
        rtf_package_bytes = Path(self.filename).read_bytes()
        package_data = self.extract_package_from_rtf(rtf_package_bytes)

        for idx, (fname, data) in enumerate(package_data.items()):
            self.objdata[fname] = data

        self.num_objdata = len(self.objdata)

    def extract_package_from_rtf(self, rtf_bytes: bytes) -> dict[str, bytes]:
        # extract file name and data from RTF Package
        fname_start = 0x22
        pos = rtf_bytes[fname_start:].find(b"\x00")
        if pos == -1:
            return {}

        fname_end = fname_start + pos
        fname = rtf_bytes[fname_start:fname_end].decode("utf-8", errors="ignore")

        path_name_start = fname_end + 1
        pos = rtf_bytes[path_name_start:].find(b"\x00")
        if pos == -1:
            return {}

        path_name_end = path_name_start + pos
        path_name = rtf_bytes[path_name_start:path_name_end].decode("utf-8", errors="ignore")

        path_name2_len = kavutil.get_uint32(rtf_bytes, path_name_end + 5)
        data_len_off = path_name_end + 5 + path_name2_len + 4
        data_len = kavutil.get_uint32(rtf_bytes, data_len_off)

        if self.verbose:
            print(f"[+] Extracted filename: {fname}")
            print(f"[+] Extracted path name: {path_name}")

        return {
            fname: rtf_bytes[data_len_off + 4 : data_len_off + 4 + data_len],
        }

    def close(self):
        pass

    def namelist(self):
        names = []

        if len(self.objdata):
            names = list(self.objdata.keys())
            names.sort()

        return names

    def read(self, fname):
        return self.objdata.get(fname, None)


# -------------------------------------------------------------------------
# RtfOle2link class
# -------------------------------------------------------------------------
class RtfOle2link:
    def __init__(self, filename, verbose=False):
        self.verbose = verbose  # Debugging mode
        self.filename = filename

        self.objdata = {}  # objdata
        self.payloads = []  # Extracted payloads
        self.parse()

    def parse(self):
        rtf_ole2link_bytes = Path(self.filename).read_bytes()
        ole2link_data = self.extract_ole2link_from_rtf(rtf_ole2link_bytes)

        for idx, (fname, data) in enumerate(ole2link_data.items()):
            self.objdata[fname] = data

        self.num_objdata = len(self.objdata)

    def extract_ole2link_from_rtf(self, rtf_bytes: bytes) -> dict[str, bytes]:
        # extract OLE2Link data from RTF OLE2Link
        if rtf_bytes.find(b"OLE2Link") != 0xC:
            return {}

        ole2link_data_len = kavutil.get_uint32(rtf_bytes, 0x1D)
        ole2link_data = rtf_bytes[0x21 : 0x21 + ole2link_data_len]

        return {
            "OLE2Link": ole2link_data,
        }

    def close(self):
        pass

    def namelist(self):
        names = []

        if len(self.objdata):
            names = list(self.objdata.keys())
            names.sort()

        return names

    def read(self, fname):
        return self.objdata.get(fname, None)


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """RTF malware detector and archive handler plugin.

    This plugin provides functionality for:
    - Detecting RTF format
    - Scanning for RTF-based exploits (CVE-2010-3333, CVE-2014-1761)
    - Extracting objects from RTF files
    """

    def __init__(self):
        """Initialize the RTF plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.1",
            title="RTF Engine",
            kmd_name="rtf",
        )
        self.cve_2010_3333_magic = None
        self.prog_cve_2010_3333_1 = None
        self.prog_cve_2010_3333_2 = None
        self.prog_cve_2014_1761 = None
        self.prog_eps_dropper = None

    def _custom_init(self) -> int:
        """Custom initialization for RTF plugin.

        Returns:
            0 for success
        """
        self.cve_2010_3333_magic = re.compile(rb"\bpfragments\b", re.IGNORECASE)
        self.prog_cve_2010_3333_1 = re.compile(
            rb"pfragments\b.*?\\sv\b.*?(\d+)|\\sv\b.*?(\d+).*?pfragments\b",
            re.IGNORECASE | re.DOTALL,
        )
        self.prog_cve_2010_3333_2 = re.compile(rb"\\sn[\W]{1,20}?pfragments\b", re.IGNORECASE)
        self.prog_cve_2014_1761 = re.compile(rb"\\listoverridecount(\d+)", re.IGNORECASE)
        self.prog_eps_dropper = re.compile(rb"exec\s+(4d5a)?([0-9a-f]{2})+50450000", re.IGNORECASE)
        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["make_arc_type"] = kernel.MASTER_DELETE
        info["sig_num"] = len(self.listvirus())
        return info

    def listvirus(self):
        """Get list of detectable viruses.

        Returns:
            List of virus names
        """
        return [
            "Exploit.RTF.CVE-2010-3333.a",
            "Exploit.RTF.CVE-2010-3333.b",
            "Exploit.RTF.CVE-2014-1761",
            "Trojan.PS.Agent",
        ]

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect RTF format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or empty dict if not recognized
        """
        ret = {}

        try:
            mm = filehandle

            if mm[:4] == b"{\\rt":
                ret["ff_rtf"] = "RTF"
            elif mm[:8] == b"\x01\x05\x00\x00\x02\x00\x00\x00":
                ret["ff_rtf_package"] = "RTF Package"
            elif mm[:8] == b"\x00\x00\x00\x00\x02\x00\x00\x00":
                ret["ff_rtf_ole2link"] = "RTF OLE2Link"

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return ret

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        try:
            mm = filehandle

            if "ff_rtf" in fileformat:
                # Check CVE-2010-3333
                if self.cve_2010_3333_magic and self.cve_2010_3333_magic.search(mm):
                    if t := self.prog_cve_2010_3333_1.search(mm):
                        val1 = int(t.groups()[0]) if t.groups()[0] else 0
                        val2 = int(t.groups()[1]) if t.groups()[1] else 0
                        val = max(val1, val2)

                        if val not in [2, 4, 8]:
                            if self.verbose:
                                print("[*] RTF :", val)
                            return True, "Exploit.RTF.CVE-2010-3333.a", 0, kernel.INFECTED

                    if self.prog_cve_2010_3333_2 and self.prog_cve_2010_3333_2.search(mm):
                        return True, "Exploit.RTF.CVE-2010-3333.b", 0, kernel.INFECTED

                # Check CVE-2014-1761
                if self.prog_cve_2014_1761:
                    if t := self.prog_cve_2014_1761.search(mm):
                        val = int(t.groups()[0])

                        if self.verbose:
                            print("[*] RTF :", val)

                        if val >= 25:
                            if t1 := re.findall(r"{\\lfolevel}", mm):
                                if self.verbose:
                                    print("[*] N :", len(t1))
                                if len(t1) > val:
                                    return True, "Exploit.RTF.CVE-2014-1761", 0, kernel.INFECTED

            elif kavutil.is_textfile(mm[:4096]):
                if self.prog_eps_dropper and self.prog_eps_dropper.search(mm):
                    return True, "Trojan.PS.Agent", 0, kernel.INFECTED

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        return False, "", -1, kernel.NOT_FOUND

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        try:
            if malware_id == 0:
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)

        return False

    def __get_handle(self, filename, handler_class):
        """Get or create handle for RTF file.

        Args:
            filename: Path to RTF file
            handler_class: Class to use for parsing

        Returns:
            Handler object or None
        """
        if filename in self.handle:
            return self.handle.get(filename, None)

        try:
            zfile = handler_class(filename, self.verbose)
            self.handle[filename] = zfile
            return zfile

        except (IOError, OSError) as e:
            logger.debug("Failed to open RTF file %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error opening RTF file %s: %s", filename, e)

        return None

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive.

        Args:
            filename: Path to RTF file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        try:
            if "ff_rtf" in fileformat:
                zfile = self.__get_handle(filename, RtfFile)
                if zfile:
                    file_scan_list.extend(["arc_rtf", name] for name in zfile.namelist())
            elif "ff_rtf_package" in fileformat:
                zfile = self.__get_handle(filename, RtfPackage)
                if zfile:
                    file_scan_list.extend(["arc_rtf_package", name] for name in zfile.namelist())
            elif "ff_rtf_ole2link" in fileformat:
                zfile = self.__get_handle(filename, RtfOle2link)
                if zfile:
                    file_scan_list.extend(["arc_rtf_ole2link", name] for name in zfile.namelist())

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive.

        Args:
            arc_engine_id: Engine ID ('arc_rtf', 'arc_rtf_package', 'arc_rtf_ole2link')
            arc_name: Path to RTF file
            fname_in_arc: Name of object to extract

        Returns:
            Extracted object data, or None on error
        """
        try:
            if arc_engine_id == "arc_rtf":
                zfile = self.__get_handle(arc_name, RtfFile)
            elif arc_engine_id == "arc_rtf_package":
                zfile = self.__get_handle(arc_name, RtfPackage)
            elif arc_engine_id == "arc_rtf_ole2link":
                zfile = self.__get_handle(arc_name, RtfOle2link)
            else:
                return None

            if zfile:
                return zfile.read(fname_in_arc)

        except (IOError, OSError) as e:
            logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        for fname in list(self.handle.keys()):
            try:
                zfile = self.handle.get(fname)
                if zfile:
                    zfile.close()
            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create an archive.

        Args:
            arc_engine_id: Engine ID
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        return arc_engine_id in ("arc_rtf", "arc_rtf_package", "arc_rtf_ole2link")



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
Macro Engine Plugin

This plugin handles Office macro malware detection and disinfection.
"""

import contextlib
import logging
import os
import re
import shutil
import struct
import tempfile
import zlib
from ctypes import POINTER, Structure, addressof, c_char, c_ubyte, c_uint, c_ushort, c_void_p, memmove, sizeof

from kicomav.plugins import cryptolib
from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.plugins import ole
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import MalwareDetectorBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# Malware ID
# -------------------------------------------------------------------------
MALWARE_ID_WORD95 = 0
MALWARE_ID_WORD97 = 1
MALWARE_ID_EXCEL95 = 2
MALWARE_ID_EXCEL97 = 3
MALWARE_ID_EXCEL_FORMULA95 = 4
MALWARE_ID_EXCEL_FORMULA97 = 5
MALWARE_ID_OLE = 99


# -------------------------------------------------------------------------
# Define the structure of Word95 macro
# -------------------------------------------------------------------------
BYTE = c_ubyte
WORD = c_ushort
DWORD = c_uint
FLOAT = c_uint
LPBYTE = POINTER(c_ubyte)
LPTSTR = POINTER(c_char)
HANDLE = c_void_p
PVOID = c_void_p
LPVOID = c_void_p
UINT_PTR = c_uint
SIZE_T = c_uint


class MCD(Structure):
    _pack_ = 1
    _fields_ = [
        ("cmg", BYTE),
        ("bEncrypt", BYTE),
        ("ibst", WORD),
        ("ibsName", WORD),
        ("ibstMenuHelp", WORD),
        ("fcFirst", DWORD),
        ("dfc", DWORD),
        ("fn", DWORD),
        ("fcFirstSave", DWORD),
    ]


class Error(Exception):
    pass


# ---------------------------------------------------------------------
# Read data
# ---------------------------------------------------------------------
def get_uint16(t_data, off):
    return struct.unpack("<H", t_data[off : off + 2])[0]


def get_uint32(t_data, off):
    return struct.unpack("<L", t_data[off : off + 4])[0]


def get_biff_record_size16(t_data, off):
    l = get_uint16(t_data, off + 2)
    return t_data[off : off + 4 + l]


def get_record_size32(t_data, off):
    l = get_uint32(t_data, off + 2)
    return t_data[off : off + 6 + l]


# ---------------------------------------------------------------------
# Create data
# ---------------------------------------------------------------------
def set_uint16(val):
    return struct.pack("<H", val)


def set_uint32(val):
    return struct.pack("<L", val)


# ---------------------------------------------------------------------
# Disinfect Word97.Macro
# ---------------------------------------------------------------------
def disinfect_word97_macro(data, verbose=False):
    if data[0x0B] & 0x01 == 0x01:
        if verbose:
            print("PASSWORD")
        return False, None  # The document is password protected

    data = data[:0x15E] + b"\x00\x00\x00\x00" + data[0x162:]
    return True, data


# ---------------------------------------------------------------------
# Disinfect Excel95.Macro
# ---------------------------------------------------------------------
def disinfect_excel95_macro(data: bytes):
    d1 = get_biff_record_size16(data, 0)
    d2 = get_biff_record_size16(data, len(d1))
    d3 = get_biff_record_size16(data, len(d1) + len(d2))

    if get_uint16(d2, 0) == 0x002F or get_uint16(d3, 0) == 0x002F:
        return False, None  # The sheet is password protected

    # Disinfect the macro in the book
    off = 0

    while off < len(data):
        t_data = get_biff_record_size16(data, off)
        val = get_uint16(t_data, 0)
        size = get_uint16(t_data, 2)

        # Manipulate the BIFF record
        if val == 0x0085:  # Sheet
            worksheet_pos = get_uint32(t_data, 4)
            if get_uint16(data, worksheet_pos + 6) == 0x06:  # Sheet containing macro : dt
                t_ret, data = disinfect_excel_worksheet(data, worksheet_pos)
                if t_ret is False:
                    return False, None

                data = data[: off + 8] + set_uint16(0x0001) + data[off + 10 :]

                len_sheet_name = data[off + 10]
                data = data[: off + 11] + (b" " * len_sheet_name) + data[off + 11 + len_sheet_name :]
        elif val in [0x00D3, 0x01BA]:  # ObProj or CodeName
            data = data[:off] + set_uint16(0x22) + data[off + 2 :]
        elif val in [0x000A, 0x0000]:  # EOF
            break

        off += len(t_data)

    return True, data


# ---------------------------------------------------------------------
# Disinfect Word95.Macro
# ---------------------------------------------------------------------
def disinfect_word95_macro(data: bytes):
    # Check if it is Word95
    start_macro = get_uint32(data, 0x118)
    if start_macro > len(data) or data[start_macro] != 0xFF:
        return False, None

    # Delete the existence of macro
    macro_magic = data[0xA] & 0xFE
    data = data[:0xA] + bytes([macro_magic]) + data[0xA + 1 :]
    data = data[:0x11C] + b"\x02\x00\x00\x00" + data[0x11C + 4 :]

    off = start_macro + 1

    while True:
        cmd = data[off]
        off += 1

        if cmd == 0x01:  # chHplmcd
            num_macro = get_uint16(data, off)
            off += 2
            break
        elif cmd == 0x02:  # chHplacd
            off += (get_uint16(data, off) * 0x04) + 2
        elif cmd in {0x03, 0x04}:  # chHplkme or chHplkmeBad
            off += (get_uint16(data, off) * 0x0E) + 2
        elif cmd == 0x05:  # chHplmud
            off += (get_uint16(data, off) * 0x0C) + 2
        elif cmd == 0x10:  # chHsttb
            off += get_uint16(data, off)
        elif cmd == 0x11:  # chMacroNames
            k = get_uint16(data, off)
            off += 2

            for _ in range(k):
                off += 2
                off += data[off] + 2
        elif cmd == 0x40:  # chTcgEnd
            return False, None

    mcd = []
    for _ in range(num_macro):
        t = MCD()
        memmove(addressof(t), data[off:], sizeof(t))
        t.fn = 0
        data = data[:off] + bytes(t) + data[off + sizeof(t) :]
        mcd.append(t)
        off += sizeof(t)

    for macro in mcd:
        t_size = macro.dfc
        t_off = macro.fcFirstSave
        data = data[:t_off] + (b"\x00" * t_size) + data[t_off + t_size :]

    data = data[: start_macro + 1] + b"\x40" + data[start_macro + 2 :]

    return True, data


# ---------------------------------------------------------------------
# Disinfect Excel97.Macro
# ---------------------------------------------------------------------
def disinfect_excel97_macro(data: bytes, verbose=False):
    d1 = get_biff_record_size16(data, 0)
    d2 = get_biff_record_size16(data, len(d1))
    d3 = get_biff_record_size16(data, len(d1) + len(d2))

    if get_uint16(d2, 0) == 0x002F or get_uint16(d3, 0) == 0x002F:
        if verbose:
            print("PASSWORD")
        return False, None  # The sheet is password protected

    # Disinfect the macro in the workbook
    off = 0

    while off < len(data):
        t_data = get_biff_record_size16(data, off)
        val = get_uint16(t_data, 0)
        size = get_uint16(t_data, 2)

        # Manipulate the BIFF record
        if val == 0x0085:  # Sheet
            worksheet_pos = get_uint32(t_data, 4)
            if get_uint16(data, worksheet_pos + 6) == 0x06:  # Sheet containing macro : dt
                t_ret, data = disinfect_excel_worksheet(data, worksheet_pos)
                if t_ret is False:
                    if verbose:
                        print("SHEET")
                    return False, None

                data = data[: off + 8] + set_uint16(0x0001) + data[off + 10 :]
                data = data[: off + 12] + (b" " * (size - 8)) + data[off + 12 + (size - 8) :]
        elif val in [0x00D3, 0x01BA]:  # ObProj or CodeName
            data = data[:off] + set_uint16(0x22) + data[off + 2 :]
        elif val in [0x000A, 0x0000]:  # EOF
            break

        off += len(t_data)

    return True, data


# ---------------------------------------------------------------------
# Disinfect Excel97.Formula
# ---------------------------------------------------------------------
def disinfect_excel97_formula(data: bytes):
    d1 = get_biff_record_size16(data, 0)
    d2 = get_biff_record_size16(data, len(d1))
    d3 = get_biff_record_size16(data, len(d1) + len(d2))

    if get_uint16(d2, 0) == 0x002F or get_uint16(d3, 0) == 0x002F:
        return False, None  # The sheet is password protected

    # Disinfect the formula in the workbook
    off = 0

    while off < len(data):
        t_data = get_biff_record_size16(data, off)
        val = get_uint16(t_data, 0)
        size = get_uint16(t_data, 2)

        # Manipulate the BIFF record
        if val == 0x0085:  # Sheet
            worksheet_pos = get_uint32(t_data, 4)
            us = get_uint16(data, off + 8)
            if us & 0x0F00 == 0x0100 or us == 0x40:
                data = data[: off + 8] + b"\x02\x00" + data[off + 10 :]
                data = data[: off + 12] + (b" " * (size - 8)) + data[off + 12 + (size - 8) :]

                t_ret, data = disinfect_excel_worksheet(data, worksheet_pos, del_formula=True)
                if t_ret is False:
                    return False, None
        elif val == 0x0018:
            data = data[: off + 4] + (b"\x00" * size) + data[off + 4 + size :]
        elif val in [0x000A, 0x0000]:  # EOF
            break

        off += len(t_data)

    return True, data


# ---------------------------------------------------------------------
# Disinfect Excel Sheet
# ---------------------------------------------------------------------
def disinfect_excel_worksheet(data: bytes, off, excel_version=MALWARE_ID_EXCEL95, del_formula=False):
    while off < len(data):
        t_data = get_biff_record_size16(data, off)
        val = get_uint16(t_data, 0)
        size = get_uint16(t_data, 2)

        if val == 0x0809:  # BOF
            data = data[: off + 6] + set_uint16(0x10) + data[off + 8 :]
        elif val == 0x023E:  # Window2
            t = set_uint16(0x08BE) + (b"\x00" * (size - 2))
            data = data[: off + 4] + t + data[off + size + 4 :]
        elif val == 0x0006:
            if del_formula:  # Disinfect the formula
                t_size = len(t_data[0x1B:]) - 2
                data = (
                    data[: off + 0x1A] + b"\x17" + set_uint16(t_size) + (b"\x00" * t_size) + data[off + 0x1D + t_size :]
                )
        elif val in [0x000A, 0x0000]:  # EOF
            break

        off += len(t_data)

    return True, data


# ---------------------------------------------------------------------
# Disinfect the macro
# ---------------------------------------------------------------------
def disinfect_office_macro(filename, malware_id):
    disinfect_macro_ref = {
        MALWARE_ID_WORD95: (None, "WordDocument", disinfect_word95_macro),
        MALWARE_ID_WORD97: ("Macros", "WordDocument", disinfect_word97_macro),
        MALWARE_ID_EXCEL95: ("_VBA_PROJECT", "Book", disinfect_excel95_macro),
        MALWARE_ID_EXCEL97: ("_VBA_PROJECT_CUR", "Workbook", disinfect_excel97_macro),
        MALWARE_ID_EXCEL_FORMULA97: (None, "Workbook", disinfect_excel97_formula),
    }

    ret = False
    while True:
        o = ole.OleFile(filename, write_mode=True)
        ole_lists = o.listdir(streams=True, storages=True)

        for name in ole_lists:
            pps = name.split("/")
            if disinfect_macro_ref[malware_id][0] and pps[-1] == disinfect_macro_ref[malware_id][0]:
                o.delete(name, delete_storage=True, reset_stream=True)
                o.close()
                break  # for
            elif pps[-1] == disinfect_macro_ref[malware_id][1]:
                pics = o.openstream(name)
                t_data = pics.read()
                t_ret, t_data = disinfect_macro_ref[malware_id][2](t_data)

                if t_ret:
                    o.write_stream(name, t_data)
                    ret = True
                else:
                    ret = False
        else:
            o.close()
            break  # while

    return ret


# ---------------------------------------------------------------------
# Analyze the dir stream
# ---------------------------------------------------------------------
def analysis_dir_stream(data, verbose=False):
    off = dir_informationrecord(data, 0, verbose)
    off = dir_referencesrecord(data, off, verbose)
    return dir_modulesrecord(data, off, verbose)


def dir_informationrecord(data, off, verbose=False):
    val = get_uint16(data, off)
    if val != 0x0001:
        raise Error("dir:InformationRecord:SysKindRecord")
    off += 10

    val = get_uint16(data, off)
    if val != 0x0002:
        raise Error("dir:InformationRecord:LcidRecord")
    off += 10

    val = get_uint16(data, off)
    if val != 0x0014:
        raise Error("dir:InformationRecord:LcidInvokeRecord")
    off += 10

    val = get_uint16(data, off)
    if val != 0x0003:
        raise Error("dir:InformationRecord:CodePageRecord")
    off += 8

    t_data = get_record_size32(data, off)
    val = get_uint16(t_data, 0)
    if val != 0x0004:
        raise Error("dir:InformationRecord:NameRecord")
    off += len(t_data)

    if verbose:
        print(f"Name : {t_data[6:]}")

    # DocStringRecord contains 2 records
    t_data = get_record_size32(data, off)
    val = get_uint16(t_data, 0)
    if val != 0x0005:
        raise Error("dir:InformationRecord:DocStringRecord #1")
    off += len(t_data)

    t_data = get_record_size32(data, off)
    val = get_uint16(t_data, 0)
    if val != 0x0040:
        raise Error("dir:InformationRecord:DocStringRecord #2")
    off += len(t_data)

    # HelpFilePathRecord contains 2 records
    t_data = get_record_size32(data, off)
    val = get_uint16(t_data, 0)
    if val != 0x0006:
        raise Error("dir:InformationRecord:HelpFilePathRecord #1")
    off += len(t_data)

    t_data = get_record_size32(data, off)
    val = get_uint16(t_data, 0)
    if val != 0x003D:
        raise Error("dir:InformationRecord:HelpFilePathRecord #2")
    off += len(t_data)

    val = get_uint16(data, off)
    if val != 0x0007:
        raise Error("dir:InformationRecord:HelpContextRecord")
    off += 10

    val = get_uint16(data, off)
    if val != 0x0008:
        raise Error("dir:InformationRecord:LibFlagsRecord")
    off += 10

    val = get_uint16(data, off)
    if val != 0x0009:
        raise Error("dir:InformationRecord:VersionRecord")
    off += 12

    # ConstantsRecord contains 2 records
    t_data = get_record_size32(data, off)
    val = get_uint16(t_data, 0)
    if val != 0x000C:
        raise Error("dir:InformationRecord:ConstantsRecord #1")
    off += len(t_data)

    t_data = get_record_size32(data, off)
    val = get_uint16(t_data, 0)
    if val != 0x003C:
        raise Error("dir:InformationRecord:ConstantsRecord #2")
    off += len(t_data)

    return off


def dir_referencesrecord(data, off, verbose=False):
    while True:
        _follow = False
        # NameRecord contains 2 records
        t_data = get_record_size32(data, off)
        val = get_uint16(t_data, 0)

        if val != 0x0016:
            break

        off += len(t_data)

        if verbose:
            print(f"ReferencesRecord Name : {t_data[6:]}")

        t_data = get_record_size32(data, off)
        val = get_uint16(t_data, 0)
        if val != 0x003E:
            raise Error("dir:ReferencesRecord:NameRecord #2")
        off += len(t_data)

        # ReferenceRecord
        val = get_uint16(data, off)
        if val == 0x0033:  # REFERENCEREGISTERED
            t_data = get_record_size32(data, off)
            off += len(t_data)
            t_data = get_record_size32(data, off)
            val = get_uint16(data, off)
            _follow = True

        if val == 0x002F:  # REFERENCECONTROL
            if not _follow:
                t_data = get_record_size32(data, off)

            off += len(t_data)
            t_data = get_record_size32(data, off)
            if get_uint16(t_data, 0) != 0x0016:
                raise Error("dir:ReferencesRecord:ReferenceRecord:REFERENCECONTROL:NameRecordExtended")

            off += len(t_data)
            t_data = get_record_size32(data, off)
            if get_uint16(t_data, 0) != 0x003E:
                raise Error("dir:ReferencesRecord:ReferenceRecord:REFERENCECONTROL:NameRecordExtended:Reserved")

            off += len(t_data)
            t_data = get_record_size32(data, off)
            if get_uint16(t_data, 0) == 0x0030:  # Reserved3
                off += len(t_data)
            else:
                raise Error("dir:ReferencesRecord:ReferenceRecord:REFERENCECONTROL:Reserved3")
        elif val in [0x000D, 0x000E]:  # REFERENCEPROJECT
            t_data = get_record_size32(data, off)
            off += len(t_data)
        else:
            raise Error("dir:ReferencesRecord:ReferenceRecord")

    return off


def dir_modulesrecord(data, off, verbose=False):
    import math

    vba_modules = []

    val = get_uint16(data, off)
    if val != 0x000F:
        raise Error("dir:ModulesRecord")
    off += 16

    while True:
        # Process the module
        t_data = get_record_size32(data, off)
        val = get_uint16(t_data, 0)
        if val != 0x0019:
            break
        off += len(t_data)

        m_name = t_data[6:]

        t_data = get_record_size32(data, off)
        val = get_uint16(t_data, 0)
        if val != 0x0047:
            continue
        off += len(t_data)

        t_data = get_record_size32(data, off)
        val = get_uint16(t_data, 0)
        if val != 0x001A:
            raise Error("dir:ModulesRecord:StreamNameRecord #1")
        off += len(t_data)

        t_data = get_record_size32(data, off)
        val = get_uint16(t_data, 0)
        if val != 0x0032:
            raise Error("dir:ModulesRecord:StreamNameRecord #2")
        off += len(t_data)

        t_data = get_record_size32(data, off)
        val = get_uint16(t_data, 0)
        if val != 0x001C:
            raise Error("dir:ModulesRecord:DocStringRecord #1")
        off += len(t_data)

        t_data = get_record_size32(data, off)
        val = get_uint16(t_data, 0)
        if val != 0x0048:
            raise Error("dir:ModulesRecord:DocStringRecord #2")
        off += len(t_data)

        t_data = get_record_size32(data, off)
        val = get_uint16(t_data, 0)
        if val != 0x0031:
            raise Error("dir:ModulesRecord:OffsetRecord")

        m_off = get_uint32(t_data, 6)
        off += 40

        if verbose:
            print("ModulesRecord Name : %s : %08X" % (m_name, m_off))

        vba_modules.append((m_name, m_off))

    return vba_modules


# ---------------------------------------------------------------------
# Decompress the macro source code in Office97
# ---------------------------------------------------------------------
def decompress(data: bytes):
    import math

    if data[0] != 1:
        return False, None

    remainder = data[1:]

    decompressed = b""
    while len(remainder) != 0:
        decompressed_chunk, remainder = decompress_chunk(remainder)

        if decompressed_chunk is None:
            return False, decompressed

        decompressed += decompressed_chunk

    return True, decompressed


def parse_token_sequence(data: bytes):
    flags = data[0]
    data = data[1:]
    result = []
    for mask in [0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80]:
        if len(data) > 0:
            if flags & mask:
                result.append(data[:2])
                data = data[2:]
            else:
                result.append(data[:1])
                data = data[1:]
    return result, data


def offset_bits(data):
    import math

    number_of_bits = int(math.ceil(math.log(len(data), 2)))
    if number_of_bits < 4:
        number_of_bits = 4
    elif number_of_bits > 12:
        number_of_bits = 12
    return number_of_bits


def decompress_chunk(compressed_chunk: bytes):
    if len(compressed_chunk) < 2:
        return None, None

    header = compressed_chunk[0] + compressed_chunk[1] * 0x100
    size = (header & 0x0FFF) + 3
    flag_compressed = header & 0x8000
    data = compressed_chunk[2 : 2 + size - 2]

    if flag_compressed == 0:
        return data, compressed_chunk[size:]

    decompressed_chunk = b""
    while len(data) != 0:
        tokens, data = parse_token_sequence(data)
        for token in tokens:
            if len(token) == 1:
                decompressed_chunk += token
            else:
                if decompressed_chunk == b"":
                    return None, None

                number_of_offset_bits = offset_bits(decompressed_chunk)
                copy_token = token[0] + token[1] * 0x100
                offset = 1 + (copy_token >> (16 - number_of_offset_bits))
                length = 3 + (((copy_token << number_of_offset_bits) & 0xFFFF) >> number_of_offset_bits)
                copy = decompressed_chunk[-offset:]
                copy = copy[:length]
                length_copy = len(copy)

                while length > length_copy:
                    if length - length_copy >= length_copy:
                        copy += copy[:length_copy]
                        length -= length_copy
                    else:
                        copy += copy[: length - length_copy]
                        length -= length - length_copy

                decompressed_chunk += copy

    return decompressed_chunk, compressed_chunk[size:]


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(MalwareDetectorBase):
    """Macro malware detector plugin.

    This plugin provides functionality for:
    - Detecting macro malware in Office documents
    - Disinfecting macro viruses
    """

    def __init__(self):
        """Initialize the Macro Engine plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="Macro Engine",
            kmd_name="macro",
        )
        self.p_vba_cmt = None
        self.p_vba_word = None
        self.word97_macro_crcs = []

    def _load_virus_database(self) -> int:
        """Load virus patterns.

        Returns:
            0 for success
        """
        self.p_vba_cmt = re.compile(rb"(\'|\bREM\b).*", re.IGNORECASE)
        self.p_vba_word = re.compile(rb"\w{2,}")

        # Suspicious Word97 macro virus CRC patterns
        self.word97_macro_crcs = [
            {0x839CE07A, 0xFC454C30, 0x084E324F},
            {0x700C7258, 0x2B408807, 0x9CECF67D},
        ]

        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        s_num = 1
        if kavutil.handle_pattern_md5:
            s_num = kavutil.handle_pattern_md5.get_sig_num("macro") + 1
        info["sig_num"] = s_num
        return info

    def listvirus(self):
        """Get list of detectable viruses.

        Returns:
            List of virus names
        """
        vlist = kavutil.handle_pattern_md5.get_sig_vlist("macro")
        if vlist is None:
            vlist = []
        vlist.append("Virus.MSExcel.Laroux.Gen")
        vlist.sort()
        return vlist

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        mm = filehandle
        o = None

        try:
            # Is there an OLE format in the pre-analyzed file format?
            if "ff_ole" in fileformat:
                o = ole.OleFile(filename)

                # Is it a vulnerable attack?
                if len(o.exploit):
                    if o:
                        o.close()
                    return True, o.exploit[0], MALWARE_ID_OLE, kernel.INFECTED

                ole_lists = o.listdir()

                for pps_name in ole_lists:
                    if pps_name.lower().find("/vba/dir") != -1:
                        pics = o.openstream(pps_name)
                        data = pics.read()
                        ret, decom_data = decompress(data)
                        if ret:
                            vba_modules = analysis_dir_stream(decom_data)

                            t = pps_name.split("/")

                            for vba in vba_modules:
                                t[-1] = vba[0].decode("utf-8", errors="ignore")
                                t_pps_name = "/".join(t)

                                t_pics = o.openstream(t_pps_name)
                                t_data = t_pics.read()

                                if len(t_data) == 0:
                                    continue

                                t_ret, buf = decompress(t_data[vba[1] :])
                                buf = buf.replace(b"\r\n", b"\n")

                                if t_ret:
                                    if self.verbose:
                                        kavutil.vprint("Macro Source")
                                        kavutil.vprint(None, "PPS", f"{t_pps_name}")
                                        print(buf)

                                    buf = self.p_vba_cmt.sub(b"", buf)
                                    buf = buf.lower()

                                    key_words = self.p_vba_word.findall(buf)

                                    vba_keyword_crc32 = set()
                                    for i in range(len(key_words) - 1):
                                        word = key_words[i] + key_words[i + 1]
                                        c = zlib.crc32(word) & 0xFFFFFFFF
                                        vba_keyword_crc32.add(c)

                                    if self.verbose:
                                        self._print_debug_keywords(key_words)

                                    # Heuristic scan
                                    for macro_crc in self.word97_macro_crcs:
                                        if macro_crc.issubset(vba_keyword_crc32):
                                            if o:
                                                o.close()
                                            return (
                                                True,
                                                "Virus.MSWord.Generic",
                                                MALWARE_ID_WORD97,
                                                kernel.SUSPECT,
                                            )

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except ole.Error as e:
            logger.debug("OLE error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        if o:
            o.close()

        return False, "", -1, kernel.NOT_FOUND

    def _print_debug_keywords(self, key_words):
        """Print debug keyword information in verbose mode."""
        max_len = len(key_words[0])

        t_word = []
        for i in range(len(key_words) - 1):
            word = key_words[i] + key_words[i + 1]
            c = zlib.crc32(word) & 0xFFFFFFFF
            t_word.append([c, key_words[i], key_words[i + 1]])

            if len(key_words[i + 1]) > max_len:
                max_len = len(key_words[i + 1])

        t_l = "+-" + ("-" * 8) + "-+-" + ("-" * max_len) + "-+-" + ("-" * max_len) + "-+"
        print(t_l)
        msg = "| %%-8s | %%-%ds | %%-%ds |" % (max_len, max_len)
        print(msg % ("CRC32", "Keyword #1", "Keyword #2"))
        print(t_l)

        msg = "| %%08X | %%-%ds | %%-%ds |" % (max_len, max_len)
        for n in t_word:
            print(msg % (n[0], n[1], n[2]))

        print(t_l)

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        # If it is an OLE vulnerability, delete it immediately
        if malware_id == MALWARE_ID_OLE:
            try:
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True
            except (IOError, OSError, k2security.SecurityError) as e:
                logger.debug("Disinfect error for %s: %s", filename, e)
                return False

        # Create a temporary file for disinfection failure recovery
        fd, t_name = tempfile.mkstemp(suffix="k2ole")
        os.close(fd)
        shutil.copy(filename, t_name)

        try:
            if malware_id == MALWARE_ID_EXCEL95:
                ret = disinfect_office_macro(filename, MALWARE_ID_EXCEL95)
            elif malware_id == MALWARE_ID_WORD95:
                ret = disinfect_office_macro(filename, MALWARE_ID_WORD95)
            elif malware_id == MALWARE_ID_EXCEL97:
                ret = disinfect_office_macro(filename, MALWARE_ID_EXCEL97)
            elif malware_id == MALWARE_ID_WORD97:
                ret = disinfect_office_macro(filename, MALWARE_ID_WORD97)
            elif malware_id == MALWARE_ID_EXCEL_FORMULA97:
                ret = disinfect_office_macro(filename, MALWARE_ID_EXCEL_FORMULA97)
            else:
                ret = False

            if ret:
                # Disinfect success, delete the temporary file
                with contextlib.suppress(k2security.SecurityError):
                    t_name_dir = os.path.dirname(t_name) or os.getcwd()
                    k2security.safe_remove_file(t_name, t_name_dir)
            else:
                # If it fails, restore the temporary file
                shutil.move(t_name, filename)

            return ret

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
            # Restore the temporary file on error
            with contextlib.suppress(Exception):
                shutil.move(t_name, filename)
            return False
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)
            return False
